#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Hong-Zhou Ye <hzyechem@gmail.com>
#

r'''
Range-separated Gaussian Density Fitting (RSGDF)

rsdf_builder.py uses a different algorithm to build RS-GDF tensors. Note both
modules CANNOT handle the long-range operator erf(omega*r12)/r12. It has to be
computed with the gdf_builder module.

ref.:
[1] For the RSGDF method:
    Hong-Zhou Ye and Timothy C. Berkelbach, J. Chem. Phys. 154, 131104 (2021).
[2] For the SR lattice sum integral screening:
    Hong-Zhou Ye and Timothy C. Berkelbach, arXiv:2107.09704.

In RSGDF, the two-center and three-center Coulomb integrals are calculated in two pars:
    j2c = j2c_SR(omega) + j2c_LR(omega)
    j3c = j3c_SR(omega) + j3c_LR(omega)
where the SR and LR integrals correspond to using the following potentials
    g_SR(r_12;omega) = erfc(omega * r_12) / r_12
    g_LR(r_12;omega) = erf(omega * r_12) / r_12
The SR integrals are evaluated in real space using a lattice summation, while
the LR integrals are evaluated in reciprocal space with a plane wave basis.
'''

import os
import h5py
import scipy.linalg
import tempfile
import numpy as np

from pyscf import lib
from pyscf.lib import logger, zdotCN
from pyscf.lib import parameters as param
from pyscf.pbc.df.df import GDF
from pyscf.pbc.df import aft, aft_jk
from pyscf.pbc.df import ft_ao
from pyscf.pbc.df import rsdf_helper
from pyscf.pbc.df import rsdf_builder
from pyscf.pbc.df import gdf_builder
from pyscf.pbc.df.incore import Int3cBuilder
from pyscf.df.outcore import _guess_shell_ranges
from pyscf.pbc.tools import k2gamma
from pyscf.pbc.lib.kpts_helper import (is_zero, member, unique,
                                       members_with_wrap_around)
from pyscf.df.addons import make_auxmol


def get_aux_chg(auxcell):
    r""" Compute charge of the auxiliary basis, \int_Omega dr chi_P(r)

    Returns:
        The function returns a 1d numpy array of size auxcell.nao_nr().
    """
    G0 = np.zeros((1, 3))
    return ft_ao.ft_ao(auxcell, G0)[0].real


class RSGDF(GDF):
    '''Range Separated Gaussian Density Fitting
    '''
    _keys = {
        'use_bvk', 'precision_R', 'precision_G', 'npw_max', '_omega_min',
        'omega', 'ke_cutoff', 'mesh_compact', 'omega_j2c', 'mesh_j2c',
        'precision_j2c', 'j2c_eig_always', 'kpts',
    }

    def weighted_coulG(self, kpt=np.zeros(3), exx=False, mesh=None, omega=None):
        return aft.weighted_coulG(self, kpt, exx, mesh, omega)

    def __init__(self, cell, kpts=np.zeros((1,3))):
        if cell.dimension < 3:
            raise NotImplementedError("""
RSGDF for low-dimensional systems are not available yet. We recommend using
cell.dimension=3 with large vacuum.""")

        # if True and kpts are gamma-inclusive, RSDF will use the bvk cell
        # trick for computing both j3c_SR and j3c_LR. If kpts are not
        # gamma-inclusive, this attribute will be ignored.
        self.use_bvk = True

        # precision for real-space lattice sum (R) and reciprocal-space
        # Fourier transform (G).
        self.precision_R = cell.precision * 1e-2
        self.precision_G = cell.precision

        # omega and PW mesh size for j3c.
        # 1. If 'omega' is given, the code can search an appropriate PW mesh of
        # size 'mesh_compact' that computes the LR-AFT of j3c to 'precision_G'.
        # 2. If 'omega' is not given, the code will search for the maximum
        # omega such that the size of 'mesh_compact' does not exceed 'npw_max'.
        # The default for 'npw_max' is 350 (i.e., 7x7x7 for a 3D cubic
        # lattice). If thus determined 'omega' is smaller than '_omega_min'
        # (default: 0.3), 'omega' will be set to '_omega_min' and 'mesh_compact'
        # is determined from the new 'omega' (ignoring 'npw_max').
        # Note 1: In both cases, the user can manually overwrite the
        # auto-determined 'mesh_compact'.
        # Note 2: 'ke_cutoff' is not an input option. Use 'mesh_compact' directly.
        self.npw_max = 350
        self._omega_min = 0.3
        self.omega = None
        self.ke_cutoff = None
        self.mesh_compact = None

        # omega and PW mesh size for j2c.
        # Like for j3c, if 'omega_j2c' is given, the code can determine an
        # appropriate PW mesh of size 'mesh_j2c' that computes the LR-AFT of j2c
        # to 'precision_j2c'.
        # The default ('omega_j2c' = 0.4 and 'precision_j2c' = 1e-14) is recommended.
        # Like for j3c, 'mesh_j2c' can be overwritten manually.
        self.omega_j2c = 0.4
        self.mesh_j2c = None
        self.precision_j2c = 1e-14

        # set True to force calculating j2c^(-1/2) using eigenvalue
        # decomposition (ED); otherwise, Cholesky decomposition (CD) is used
        # first, and ED is called only if CD fails.
        self.j2c_eig_always = False

        GDF.__init__(self, cell, kpts=kpts)

        self.kpts = np.reshape(self.kpts, (-1,3))

    def dump_flags(self, verbose=None):
        cell = self.cell
        log = logger.new_logger(self, verbose)
        log.info('\n')
        log.info('******** %s ********', self.__class__)
        log.info('cell num shells = %d, num cGTOs = %d, num pGTOs = %d',
                 cell.nbas, cell.nao_nr(), cell.npgto_nr())
        log.info('use_bvk = %s', self.use_bvk)
        log.info('precision_R = %s', self.precision_R)
        log.info('precision_G = %s', self.precision_G)
        log.info('j2c_eig_always = %s', self.j2c_eig_always)
        log.info('omega = %s', self.omega)
        log.info('ke_cutoff = %s', self.ke_cutoff)
        if self.mesh is not None:
            log.info('mesh = %s (%d PWs)', self.mesh, np.prod(self.mesh))
        log.info('mesh_compact = %s (%d PWs)', self.mesh_compact,
                 np.prod(self.mesh_compact))
        if self.auxcell is None:
            log.info('auxbasis = %s', self.auxbasis)
        else:
            log.info('auxbasis = %s', self.auxcell.basis)
            log.info('auxcell precision= %s', self.auxcell.precision)
            log.info('auxcell rcut = %s', self.auxcell.rcut)
            log.info('omega_j2c = %s', self.omega_j2c)
            log.info('mesh_j2c = %s (%d PWs)', self.mesh_j2c,
                     np.prod(self.mesh_j2c))

        auxcell = self.auxcell
        log.info('auxcell num shells = %d, num cGTOs = %d, num pGTOs = %d',
                 auxcell.nbas, auxcell.nao_nr(),
                 auxcell.npgto_nr())

        log.info('exp_to_discard = %s', self.exp_to_discard)
        if isinstance(self._cderi, str):
            log.info('_cderi = %s  where DF integrals are loaded (readonly).',
                     self._cderi)
        elif isinstance(self._cderi_to_save, str):
            log.info('_cderi_to_save = %s', self._cderi_to_save)
        else:
            log.info('_cderi_to_save = %s', self._cderi_to_save.name)
        log.info('len(kpts) = %d', len(self.kpts))
        log.debug1('    kpts = %s', self.kpts)
        if self.kpts_band is not None:
            log.info('len(kpts_band) = %d', len(self.kpts_band))
            log.debug1('    kpts_band = %s', self.kpts_band)

        return self

    def _rs_build(self):
        log = logger.Logger(self.stdout, self.verbose)

        # find kmax
        kpts = self.kpts if self.kpts_band is None else np.vstack(
                                                    [self.kpts, self.kpts_band])
        b = self.cell.reciprocal_vectors()
        scaled_kpts = np.linalg.solve(b.T, kpts.T).T
        scaled_kpts[scaled_kpts > 0.49999999] -= 1
        kpts = np.dot(scaled_kpts, b)
        kmax = np.linalg.norm(kpts, axis=-1).max()
        scaled_kpts = kpts = None
        if kmax < 1.e-3: kmax = (0.75/np.pi/self.cell.vol)**0.33333333*2*np.pi

        # If omega is not given, estimate it from npw_max
        r2o = True
        if self.omega is None:
            self.omega, self.ke_cutoff, mesh_compact = \
                                rsdf_helper.estimate_omega_for_npw(
                                                self.cell, self.npw_max,
                                                self.precision_G,
                                                kmax=kmax,
                                                round2odd=r2o)
            # if omega from npw_max is too small, use omega_min
            if self.omega < self._omega_min:
                self.omega = self._omega_min
                self.ke_cutoff, mesh_compact = \
                                    rsdf_helper.estimate_mesh_for_omega(
                                                    self.cell, self.omega,
                                                    self.precision_G,
                                                    kmax=kmax,
                                                    round2odd=r2o)
            # Use the thus determined mesh_compact only if not p[rovided
            if self.mesh_compact is None:
                self.mesh_compact = mesh_compact
        # If omega is provided but mesh_compact is not
        elif self.mesh_compact is None:
            self.ke_cutoff, self.mesh_compact = \
                                rsdf_helper.estimate_mesh_for_omega(
                                                self.cell, self.omega,
                                                self.precision_G,
                                                kmax=kmax,
                                                round2odd=r2o)

        self.mesh_compact = self.cell.symmetrize_mesh(self.mesh_compact)

        # build auxcell
        auxcell = make_auxmol(self.cell, self.auxbasis)
        # drop exponents
        drop_eta = self.exp_to_discard
        if drop_eta is not None and drop_eta > 0:
            log.info("Drop primitive fitting functions with exponent < %s",
                     drop_eta)
            auxbasis = rsdf_helper.remove_exp_basis(auxcell._basis,
                                                    amin=drop_eta)
            auxcellnew = make_auxmol(self.cell, auxbasis)
            auxcell = auxcellnew

        # determine mesh for computing j2c
        auxcell.precision = self.precision_j2c
        auxcell.rcut = max([auxcell.bas_rcut(ib, auxcell.precision)
                            for ib in range(auxcell.nbas)])
        if self.mesh_j2c is None:
            self.mesh_j2c = rsdf_helper.estimate_mesh_for_omega(
                                    auxcell, self.omega_j2c, round2odd=True)[1]
        self.mesh_j2c = self.cell.symmetrize_mesh(self.mesh_j2c)
        self.auxcell = auxcell

    def _kpts_build(self, kpts_band=None):
        if self.kpts_band is not None:
            self.kpts_band = np.reshape(self.kpts_band, (-1,3))
        if kpts_band is not None:
            kpts_band = np.reshape(kpts_band, (-1,3))
            if self.kpts_band is None:
                self.kpts_band = kpts_band
            else:
                self.kpts_band = unique(np.vstack((self.kpts_band,kpts_band)))[0]

    def _gdf_build(self, j_only=None, with_j3c=True):
        if j_only is not None:
            self._j_only = j_only

        if with_j3c:
            if isinstance(self._cderi_to_save, str):
                cderi = self._cderi_to_save
            else:
                cderi = self._cderi_to_save.name
            if isinstance(self._cderi, str):
                if self._cderi == cderi and os.path.isfile(cderi):
                    logger.warn(self, 'File %s (specified by ._cderi) is '
                                'overwritten by GDF initialization.', cderi)
                else:
                    logger.warn(self, 'Value of ._cderi is ignored. '
                                'DF integrals will be saved in file %s .', cderi)
            self._cderi = cderi
            t1 = (logger.process_clock(), logger.perf_counter())
            self._make_j3c(self.cell, self.auxcell, None, cderi)
            t1 = logger.timer_debug1(self, 'j3c', *t1)

    def _make_j3c(self, cell=None, auxcell=None, kptij_lst=None, cderi_file=None):
        if cell is None: cell = self.cell
        if auxcell is None: auxcell = self.auxcell
        if cderi_file is None: cderi_file = self._cderi_to_save

        if self.kpts_band is None:
            kpts_union = self.kpts
        else:
            kpts_union = unique(np.vstack([self.kpts, self.kpts_band]))[0]
        dfbuilder = _RSGDFBuilder(cell, auxcell, kpts_union)
        dfbuilder.__dict__.update(self.__dict__)
        dfbuilder.kpts = kpts_union
        j_only = self._j_only or len(kpts_union) == 1
        dfbuilder.make_j3c(cderi_file, j_only=j_only, dataname=self._dataname,
                           kptij_lst=kptij_lst)

    def build(self, j_only=None, with_j3c=True, kpts_band=None):
        # formatting k-points
        self._kpts_build(kpts_band=kpts_band)

        # build for range-separation hybrid
        self._rs_build()

        # dump flags before the final build
        self.check_sanity()
        self.dump_flags()

        # do normal gdf build with the modified _make_j3c
        self._gdf_build(j_only=j_only, with_j3c=with_j3c)

        return self


RSDF = RSGDF


class _RSGDFBuilder(rsdf_builder._RSGDFBuilder):
    _keys = {
        'use_bvk', 'precision_R', 'precision_G', 'npw_max', '_omega_min',
        'omega', 'ke_cutoff', 'mesh_compact', 'omega_j2c', 'mesh_j2c',
        'precision_j2c', 'j2c_eig_always', 'kpts',
    }

    def __init__(self, cell, auxcell, kpts=np.zeros((1,3))):
        self.eta = None
        self.mesh = None
        if cell.omega != 0:
            # Initialize omega to cell.omega for HF exchange of short range
            # int2e in RSH functionals
            self.omega = abs(cell.omega)
        else:
            self.omega = None
        self.ke_cutoff = None
        self.bvk_kmesh = None
        self.supmol_ft = None

        Int3cBuilder.__init__(self, cell, auxcell, kpts)

        # set True to force calculating j2c^(-1/2) using eigenvalue
        # decomposition (ED); otherwise, Cholesky decomposition (CD) is used
        # first, and ED is called only if CD fails.
        self.j2c_eig_always = False
        self.linear_dep_threshold = rsdf_builder.LINEAR_DEP_THR

    def build(self, omega=None):
        log = logger.new_logger(self)
        cell = self.cell
        kpts = self.kpts

        self.bvk_kmesh = kmesh = k2gamma.kpts_to_kmesh(cell, kpts)
        log.debug('kmesh for bvk-cell = %s', kmesh)

        self.rs_cell = rs_cell = ft_ao._RangeSeparatedCell.from_cell(
            cell, self.ke_cutoff, rsdf_builder.RCUT_THRESHOLD, verbose=log)

        rcut = rsdf_builder.estimate_ft_rcut(rs_cell, cell.precision,
                                             exclude_dd_block=False)
        supmol_ft = rsdf_builder._ExtendedMoleFT.from_cell(rs_cell, kmesh,
                                                           rcut.max(), log)
        supmol_ft.exclude_dd_block = False
        self.supmol_ft = supmol_ft.strip_basis(rcut)
        log.debug('sup-mol-ft nbas = %d cGTO = %d pGTO = %d',
                  supmol_ft.nbas, supmol_ft.nao, supmol_ft.npgto_nr())
        return self

    def get_2c2e(self, uniq_kpts):
        cell = self.cell
        auxcell = self.auxcell
        # compute j2c first as it informs the integral screening in computing j3c
        # short-range part of j2c ~ (-kpt_ji | kpt_ji)
        omega_j2c = abs(self.omega_j2c)
        j2c = rsdf_helper.intor_j2c(auxcell, omega_j2c, kpts=uniq_kpts)

        # get charge of auxbasis
        if cell.dimension == 3:
            qaux = get_aux_chg(auxcell)
        else:
            qaux = np.zeros(auxcell.nao_nr())

        # Add (1) short-range G=0 (i.e., charge) part and (2) long-range part
        qaux2 = None
        g0_j2c = np.pi/omega_j2c**2./cell.vol
        mesh_j2c = self.mesh_j2c
        Gv, Gvbase, kws = cell.get_Gv_weights(mesh_j2c)
        b = cell.reciprocal_vectors()
        gxyz = lib.cartesian_prod([np.arange(len(x)) for x in Gvbase])
        ngrids = gxyz.shape[0]

        max_memory = max(2000, self.max_memory - lib.current_memory()[0])
        blksize = max(2048, int(max_memory*.5e6/16/auxcell.nao_nr()))
        logger.debug2(self, 'max_memory %s (MB)  blocksize %s', max_memory, blksize)

        for k, kpt in enumerate(uniq_kpts):
            # short-range charge part
            if is_zero(kpt) and cell.dimension == 3:
                if qaux2 is None:
                    qaux2 = np.outer(qaux,qaux)
                j2c[k] -= qaux2 * g0_j2c
            # long-range part via aft
            coulG_lr = self.weighted_coulG(kpt, mesh=mesh_j2c, omega=omega_j2c)
            for p0, p1 in lib.prange(0, ngrids, blksize):
                auxG = ft_ao.ft_ao(auxcell, Gv[p0:p1], None, b, gxyz[p0:p1], Gvbase, kpt).T
                auxGR = np.asarray(auxG.real, order='C')
                auxGI = np.asarray(auxG.imag, order='C')
                auxG = None

                if is_zero(kpt):  # kpti == kptj
                    j2c[k] += lib.ddot(auxGR*coulG_lr[p0:p1], auxGR.T)
                    j2c[k] += lib.ddot(auxGI*coulG_lr[p0:p1], auxGI.T)
                else:
                    j2cR, j2cI = zdotCN(auxGR*coulG_lr[p0:p1],
                                        auxGI*coulG_lr[p0:p1], auxGR.T, auxGI.T)
                    j2c[k] += j2cR + j2cI * 1j
                auxGR = auxGI = j2cR = j2cI = None
        return j2c

    def outcore_auxe2(self, cderi_file, intor='int3c2e', aosym='s2', comp=None,
                      kptij_lst=None, j_only=False, dataname='j3c-junk',
                      shls_slice=None):
        # Deadlock on NFS if you open an already-opened tmpfile in H5PY
        # swapfile = tempfile.NamedTemporaryFile(dir=os.path.dirname(cderi_file))
        fswap = lib.H5TmpFile(dir=os.path.dirname(cderi_file), prefix='.outcore_auxe2_swap')
        # avoid trash files
        os.unlink(fswap.filename)

        cell = self.cell
        if self.use_bvk and self.kpts_band is None:
            bvk_kmesh = self.bvk_kmesh
        else:
            bvk_kmesh = None

        max_memory = max(2000, self.max_memory-lib.current_memory()[0])
        rsdf_helper._aux_e2_nospltbas(cell, self.auxcell, self.omega,
                                      fswap, intor, aosym=aosym,
                                      kptij_lst=kptij_lst, dataname=dataname,
                                      max_memory=max_memory,
                                      bvk_kmesh=bvk_kmesh,
                                      precision=self.precision_R)
        return fswap

    def weighted_ft_ao(self, kpt):
        cell = self.cell
        auxcell = self.auxcell
        mesh = self.mesh_compact
        Gv, Gvbase, kws = cell.get_Gv_weights(mesh)
        b = cell.reciprocal_vectors()
        gxyz = lib.cartesian_prod([np.arange(len(x)) for x in Gvbase])
        shls_slice = (0, auxcell.nbas)
        auxG = ft_ao.ft_ao(auxcell, Gv, shls_slice, b, gxyz, Gvbase, kpt).T
        wcoulG_lr = self.weighted_coulG(kpt, mesh=mesh, omega=self.omega)
        auxG *= wcoulG_lr
        Gaux = lib.transpose(auxG)
        GauxR = np.asarray(Gaux.real, order='C')
        GauxI = np.asarray(Gaux.imag, order='C')
        return GauxR, GauxI

    def gen_j3c_loader(self, h5group, kpt, kpt_ij_idx, ijlst_mapping, aosym):
        cell = self.cell
        kpts = self.kpts
        nkpts = len(self.kpts)
        vbar = None
        if is_zero(kpt) and cell.dimension == 3:
            qaux = get_aux_chg(self.auxcell)
            vbar = np.pi / self.omega**2 / cell.vol * qaux
            vbar_idx = np.where(vbar != 0)[0]
            ovlp = cell.pbc_intor('int1e_ovlp', hermi=1, kpts=kpts)
            if aosym == 's2':
                ovlp = [lib.pack_tril(s) for s in ovlp]
            else:
                ovlp = [s.ravel() for s in ovlp]

        # TODO: Store rs_density_fit cderi tensor in v1 format for the moment.
        # It should be changed to 'v2' format in the future.
        if ijlst_mapping is None:
            data_version = 'v2'
        else:
            data_version = 'v1'

        if data_version == 'v1':
            nsegs = len(h5group[f'j3c-junk/{ijlst_mapping[kpt_ij_idx[0]]}'])
        else:
            nsegs = len(h5group[f'j3c-junk/{kpt_ij_idx[0]}'])

        def load_j3c(col0, col1):
            j3cR = []
            j3cI = []
            for kk in kpt_ij_idx:
                if data_version == 'v1':
                    v = np.hstack([h5group[f'j3c-junk/{ijlst_mapping[kk]}/{i}'][0,col0:col1]
                                   for i in range(nsegs)])
                else:
                    v = np.hstack([h5group[f'j3c-junk/{kk}/{i}'][0,col0:col1]
                                   for i in range(nsegs)])
                vR = np.asarray(v.real, order='C')
                kj = kk % nkpts
                if is_zero(kpt) and is_zero(kpts[kj]):
                    vI = None
                else:
                    vI = np.asarray(v.imag, order='C')
                # vbar is the interaction between the background charge
                # and the auxiliary basis.  0D, 1D, 2D do not have vbar.
                if vbar is not None:
                    vmod = ovlp[kj][col0:col1,None] * vbar[vbar_idx]
                    vR[:,vbar_idx] -= vmod.real
                    if vI is not None:
                        vI[:,vbar_idx] -= vmod.imag
                j3cR.append(vR)
                j3cI.append(vI)
            return j3cR, j3cI

        return load_j3c

    def make_j3c(self, cderi_file, intor='int3c2e', aosym='s2', comp=None,
                 j_only=False, dataname='j3c', shls_slice=None, kptij_lst=None):
        if self.cell.omega != 0:
            raise RuntimeError('RSGDF cannot be used to evaluate the long-range '
                               'HF exchange in RSH functionals.')

        cpu1 = (logger.process_clock(), logger.perf_counter())
        log = logger.Logger(self.stdout, self.verbose)
        if self.rs_cell is None:
            self.build()

        cell = self.cell
        kpts = self.kpts
        nkpts = len(kpts)
        nao = cell.nao
        naux = self.auxcell.nao

        if shls_slice is not None:
            raise NotImplementedError
        ish0, ish1 = 0, cell.nbas

        # ijlst_mapping maps the [nkpts x nkpts] kpts-pair to kpts-pair in
        # kptij_lst. Value -1 in ijlst_mapping means the kpts-pair does not
        # exist in kptij_lst
        ijlst_mapping = np.empty(nkpts * nkpts, dtype=int)
        ijlst_mapping[:] = -1
        if kptij_lst is None:
            if j_only:
                kpti_idx = np.arange(nkpts)
                ijlst_mapping[kpti_idx * nkpts + kpti_idx] = kpti_idx
                kptij_lst = np.concatenate([kpts[:,None,:], kpts[:,None,:]], axis=1)
                kk_idx = kpti_idx * nkpts + kpti_idx
            else:
                kpti_idx, kptj_idx = np.tril_indices(nkpts)
                nkpts_pair = kpti_idx.size
                ijlst_mapping[kpti_idx * nkpts + kptj_idx] = np.arange(nkpts_pair)
                kptij_lst = np.concatenate([kpts[kpti_idx,None,:],
                                            kpts[kptj_idx,None,:]], axis=1)
                kk_idx = kpti_idx * nkpts + kptj_idx
        else:
            kpti_idx = members_with_wrap_around(cell, kptij_lst[:,0], kpts)
            kptj_idx = members_with_wrap_around(cell, kptij_lst[:,1], kpts)
            ijlst_mapping[kpti_idx * nkpts + kptj_idx] = np.arange(len(kptij_lst))
            kk_idx = kpti_idx * nkpts + kptj_idx

        # TODO: Store rs_density_fit cderi tensor in v1 format for the moment.
        # It should be changed to 'v2' format in the future.
        data_version = 'v1'
        if h5py.is_hdf5(cderi_file):
            feri = lib.H5FileWrap(cderi_file, 'a')
            if 'kpts' in feri:
                del feri['j3c-kptij']
            if dataname in feri:
                log.warn(f'Overwritting {dataname} in {cderi_file}.')
                del feri[dataname]
        else:
            feri = lib.H5FileWrap(cderi_file, 'w')
        feri['j3c-kptij'] = kptij_lst

        fswap = self.outcore_auxe2(cderi_file, intor, aosym, comp,
                                   kptij_lst, j_only, 'j3c-junk', shls_slice)
        cpu1 = log.timer_debug1('3c2e', *cpu1)

        ft_kern = self.supmol_ft.gen_ft_kernel(aosym, return_complex=False,
                                               verbose=log)

        # recompute g0 and Gvectors for j3c
        mesh = self.mesh_compact
        Gv, Gvbase, kws = cell.get_Gv_weights(mesh)
        gxyz = lib.cartesian_prod([np.arange(len(x)) for x in Gvbase])
        ngrids = gxyz.shape[0]

        # Add (1) short-range G=0 (i.e., charge) part and (2) long-range part
        tspans = np.zeros((3,2))    # lr, j2c_inv, j2c_cntr
        tspannames = ["ftaop+pw", "j2c_inv", "j2c_cntr"]
        def make_cderi(kpt, kpt_ij_idx, j2c):
            log.debug1('make_cderi for %s', kpt)
            kptjs = kpts[kpt_ij_idx % nkpts]
            nkptj = len(kptjs)
            if data_version == 'v1':
                input_kptij_idx = ijlst_mapping[kpt_ij_idx]
                # filter kpt_ij_idx, keeps only the kpts-pair in kptij_lst
                kpt_ij_idx = kpt_ij_idx[input_kptij_idx >= 0]
                # input_kptij_idx saves the indices of remaining kpts-pair in kptij_lst
                input_kptij_idx = input_kptij_idx[input_kptij_idx >= 0]
                log.debug1('kpt_ij_idx = %s', kpt_ij_idx)
                log.debug1('input_kptij_idx = %s', input_kptij_idx)
            else:
                input_kptij_idx = kpt_ij_idx
            if kpt_ij_idx.size == 0:
                return

            Gaux = self.weighted_ft_ao(kpt)

            if is_zero(kpt):  # kpti == kptj
                aosym = 's2'
                nao_pair = nao*(nao+1)//2
            else:
                aosym = 's1'
                nao_pair = nao**2

            load = self.gen_j3c_loader(fswap, kpt, kpt_ij_idx, ijlst_mapping, aosym)

            mem_now = lib.current_memory()[0]
            log.debug2('memory = %s', mem_now)
            max_memory = max(1000, self.max_memory - mem_now)
            # nkptj for 3c-coulomb arrays plus 1 Lpq array
            buflen = min(max(int(max_memory*.3e6/16/naux/(nkptj+1)), 1), nao_pair)
            sh_ranges = _guess_shell_ranges(cell, buflen, aosym, start=ish0, stop=ish1)
            buflen = max([x[2] for x in sh_ranges])
            # * 2 for the buffer used in preload
            max_memory -= buflen * naux * (nkptj+1) * 16e-6 * 2

            # +1 for a pqkbuf
            Gblksize = max(16, int(max_memory*1e6/16/buflen/(nkptj+1)))
            Gblksize = min(Gblksize, ngrids, 200000)

            cols = [sh_range[2] for sh_range in sh_ranges]
            locs = np.append(0, np.cumsum(cols))
            # buf for ft_aopair
            buf = np.empty(nkptj*buflen*Gblksize, dtype=np.complex128)
            for istep, j3c in enumerate(lib.map_with_prefetch(load, locs[:-1], locs[1:])):
                bstart, bend, ncol = sh_ranges[istep]
                log.debug1('int3c2e [%d/%d], AO [%d:%d], ncol = %d',
                           istep+1, len(sh_ranges), bstart, bend, ncol)
                if aosym == 's2':
                    shls_slice = (bstart, bend, 0, bend)
                else:
                    shls_slice = (bstart, bend, 0, cell.nbas)

                if self.has_long_range():
                    tick_ = np.asarray((logger.process_clock(), logger.perf_counter()))
                    for p0, p1 in lib.prange(0, ngrids, Gblksize):
                        # shape of Gpq (nkpts, nGv, ni, nj)
                        Gpq = ft_kern(Gv[p0:p1], gxyz[p0:p1], Gvbase, kpt,
                                      kptjs, shls_slice, aosym, out=buf)
                        self.add_ft_j3c(j3c, Gpq, Gaux, p0, p1)
                        Gpq = None
                    tock_ = np.asarray((logger.process_clock(), logger.perf_counter()))
                    tspans[0] += tock_ - tick_

                j3cR, j3cI = j3c
                for k, idx in enumerate(input_kptij_idx):
                    cderi, cderi_negative = self.solve_cderi(j2c, j3cR[k], j3cI[k])
                    feri[f'{dataname}/{idx}/{istep}'] = cderi
                    if cderi_negative is not None:
                        # for low-dimension systems
                        feri[f'{dataname}-/{idx}/{istep}'] = cderi_negative
                j3cR = j3cI = j3c = cderi = None
                tick_ = np.asarray((logger.process_clock(), logger.perf_counter()))
                tspans[2] += tick_ - tock_

        for kpt, kpt_ij_idx, cd_j2c \
                in self.gen_uniq_kpts_groups(j_only, fswap, kk_idx=kk_idx):
            make_cderi(kpt, kpt_ij_idx, cd_j2c)

        feri.close()
        # report time for aft part
        for tspan, tspanname in zip(tspans, tspannames):
            log.debug1("    CPU time for %s %9.2f sec, wall time %9.2f sec",
                       "%10s"%tspanname, *tspan)
        return self



#!/usr/bin/env python
# Copyright 2025 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Finite difference driver
'''

import numpy as np
from pyscf import gto
from pyscf import lib
from pyscf.lib import logger
from pyscf.grad.rhf import GradientsBase
from pyscf.hessian.rhf import HessianBase

def kernel(method, displacement=1e-2):
    '''
    Evaluate gradients or Hessians for a given method using finite difference approximation.

    Args:
        method (callable):
            The function for which the gradient or Hessian is to be computed.

    Kwargs:
        displacement:
            The small change for finite difference calculations. Default is 1e-2.

    Returns:
        An (n, 3) array for gradients or (n, n, 3, 3) array for hessian,
        depending on the given method.
    '''
    assert isinstance(method, lib.StreamObject)

    mol = method.mol
    original_coords = mol.atom_coords()
    natm = mol.natm
    if isinstance(method, GradientsBase):
        logger.info(mol, 'Computing finite-difference Hessian for {method}')
        if method.base.conv_tol > displacement**3:
            logger.warn(mol, 'conv_tol %g with displacement %g might be insufficinet',
                        method.base.conv_tol, displacement)
        de = np.empty((natm,3,natm,3))
    else:
        logger.info(mol, 'Computing finite-difference Gradients for {method}')
        if method.conv_tol > displacement**2:
            logger.warn(mol, 'conv_tol %g with displacement %g might be insufficinet',
                        method.conv_tol, displacement)
        de = np.empty((natm,3))

    scan = None
    if isinstance(method, (lib.SinglePointScanner, lib.GradScanner)):
        scan = method
    elif hasattr(method, 'as_scanner'):
        logger.info(mol, 'Apply {method}.scanner')
        scan = method.as_scanner()
    else:
        method = method.copy()
        if isinstance(method, GradientsBase):
            method.base = method.base.copy()

    if scan is not None:
        def evaluate(r):
            mol.set_geom_(r, unit='Bohr')
            if scan:
                if isinstance(method, GradientsBase):
                    res = scan(mol)[1]
                else:
                    res = scan(mol)
                if not scan.converged:
                    raise RuntimeError('{scan} not converged')
                return res
    else:
        logger.info(mol, '{method}.scanner not found. '
                    'Initial guess may not be utilized among different geometries')
        def evaluate(r):
            if isinstance(method, GradientsBase):
                method.base.mol.set_geom_(r, unit='Bohr')
                method.base.run()
                if not method.base.converged:
                    raise RuntimeError('{method.base} not converged')
                method.mol.set_geom_(r, unit='Bohr')
                return method.kernel()
            else:
                method.mol.set_geom_(r, unit='Bohr')
                if not method.converged:
                    raise RuntimeError('{method} not converged')
                res = method.kernel()
            return res

    try:
        atom_coords = original_coords.copy()
        for i in range(natm):
            for x in range(3):
                atom_coords[i,x] += displacement
                e1 = evaluate(atom_coords)
                atom_coords[i,x] -= 2*displacement
                e2 = evaluate(atom_coords)
                de[i,x] = (e1 - e2) / (2*displacement)
                atom_coords[i,x] = original_coords[i,x]
    finally:
        mol.set_geom_(original_coords, unit='Bohr')

    if isinstance(method, GradientsBase):
        # Hessian is stored as (N,N,3,3)
        de = de.transpose(0,2,1,3)
    return de

class Gradients(GradientsBase):
    displacement = 1e-2

    def __init__(self, method):
        assert isinstance(method, lib.StreamObject)
        assert not isinstance(method, GradientsBase)
        self.base = method
        self.mol = mol = method.mol
        self.stdout = mol.stdout
        self.verbose = mol.verbose
        self.de = None

    def kernel(self):
        self.de = kernel(self.base, self.displacement)
        return self.de

    def as_scanner(self):
        if isinstance(self, lib.GradScanner):
            return self

        logger.info(self, 'Create Gradient scanner for %s', self.base.__class__)
        name = 'FiniteDiffGrad' + GradScanner.__name_mixin__
        return lib.set_class(GradScanner(self),
                             (GradScanner, self.__class__), name)

class GradScanner(lib.GradScanner):
    def __call__(self, mol_or_geom, **kwargs):
        if isinstance(mol_or_geom, gto.MoleBase):
            assert mol_or_geom.__class__ == gto.Mole
            mol = mol_or_geom
        else:
            mol = self.mol.set_geom_(mol_or_geom, inplace=False)

        self.base(mol)
        e_tot = self.base.e_tot
        de = self.kernel()
        return e_tot, de

class Hessian(HessianBase):
    displacement = 1e-2

    def __init__(self, method):
        assert isinstance(method, lib.StreamObject)
        assert isinstance(method, GradientsBase)
        self.base = method.base
        self._method = method
        self.mol = mol = method.mol
        self.stdout = mol.stdout
        self.verbose = mol.verbose
        self.de = None

    def kernel(self):
        self.de = kernel(self._method, self.displacement)
        return self.de

    def as_scanner(self):
        return self



#!/usr/bin/env python
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Timothy Berkelbach <tim.berkelbach@gmail.com>
#         James McClain <jdmcclain47@gmail.com>
#         Xing Zhang <zhangxing.nju@gmail.com>
#


'''
kpoint-adapted and spin-adapted MP2
t2[i,j,a,b] = <ij|ab> / D_ij^ab

t2 and eris are never stored in full, only a partial
eri of size (nkpts,nocc,nocc,nvir,nvir)
'''

import numpy as np
from scipy.linalg import block_diag
import h5py

from pyscf import lib
from pyscf.lib import logger, einsum
from pyscf.mp import mp2
from pyscf.pbc.df import df
from pyscf.pbc.lib import kpts_helper
from pyscf.pbc.lib import kpts as libkpts
from pyscf.lib.parameters import LARGE_DENOM
from pyscf import __config__

WITH_T2 = getattr(__config__, 'mp_mp2_with_t2', True)

def kernel(mp, mo_energy, mo_coeff, verbose=logger.NOTE, with_t2=WITH_T2):
    """Computes k-point RMP2 energy.

    Args:
        mp (KMP2): an instance of KMP2
        mo_energy (list): a list of numpy.ndarray. Each array contains MO energies of
                          shape (Nmo,) for one kpt
        mo_coeff (list): a list of numpy.ndarray. Each array contains MO coefficients
                         of shape (Nao, Nmo) for one kpt
        verbose (int, optional): level of verbosity. Defaults to logger.NOTE (=3).
        with_t2 (bool, optional): whether to compute t2 amplitudes. Defaults to WITH_T2 (=True).

    Returns:
        KMP2 energy and t2 amplitudes (=None if with_t2 is False)
    """
    cput0 = (logger.process_clock(), logger.perf_counter())
    log = logger.new_logger(mp, verbose)

    mp.dump_flags()
    nmo = mp.nmo
    nocc = mp.nocc
    nvir = nmo - nocc
    nkpts = mp.nkpts

    with_df_ints = mp.with_df_ints and isinstance(mp._scf.with_df, df.GDF)

    mem_avail = mp.max_memory - lib.current_memory()[0]
    mem_usage = (nkpts * (nocc * nvir)**2) * 16 / 1e6
    if with_df_ints:
        mydf = mp._scf.with_df
        if mydf.auxcell is None:
            # Calculate naux based on precomputed GDF integrals
            naux = mydf.get_naoaux()
        else:
            naux = mydf.auxcell.nao_nr()

        mem_usage += (nkpts**2 * naux * nocc * nvir) * 16 / 1e6
    if with_t2:
        mem_usage += (nkpts**3 * (nocc * nvir)**2) * 16 / 1e6
    if mem_usage > mem_avail:
        raise MemoryError('Insufficient memory! MP2 memory usage %d MB (currently available %d MB)'
                          % (mem_usage, mem_avail))

    eia = np.zeros((nocc,nvir))
    eijab = np.zeros((nocc,nocc,nvir,nvir))

    fao2mo = mp._scf.with_df.ao2mo
    kconserv = mp.khelper.kconserv
    oovv_ij = np.zeros((nkpts,nocc,nocc,nvir,nvir), dtype=mo_coeff[0].dtype)

    mo_e_o = [mo_energy[k][:nocc] for k in range(nkpts)]
    mo_e_v = [mo_energy[k][nocc:] for k in range(nkpts)]

    # Get location of non-zero/padded elements in occupied and virtual space
    nonzero_opadding, nonzero_vpadding = padding_k_idx(mp, kind="split")

    if with_t2:
        t2 = np.zeros((nkpts, nkpts, nkpts, nocc, nocc, nvir, nvir), dtype=complex)
    else:
        t2 = None

    # Build 3-index DF tensor Lov
    if with_df_ints:
        Lov = _init_mp_df_eris(mp)

    emp2_ss = emp2_os = 0.
    for ki in range(nkpts):
        for kj in range(nkpts):
            for ka in range(nkpts):
                kb = kconserv[ki,ka,kj]
                # (ia|jb)
                if with_df_ints:
                    oovv_ij[ka] = (1./nkpts) * einsum("Lia,Ljb->iajb", Lov[ki, ka], Lov[kj, kb]).transpose(0,2,1,3)
                else:
                    orbo_i = mo_coeff[ki][:,:nocc]
                    orbo_j = mo_coeff[kj][:,:nocc]
                    orbv_a = mo_coeff[ka][:,nocc:]
                    orbv_b = mo_coeff[kb][:,nocc:]
                    oovv_ij[ka] = fao2mo((orbo_i,orbv_a,orbo_j,orbv_b),
                                         (mp.kpts[ki],mp.kpts[ka],mp.kpts[kj],mp.kpts[kb]),
                                         compact=False).reshape(nocc,nvir,nocc,nvir).transpose(0,2,1,3) / nkpts
            for ka in range(nkpts):
                kb = kconserv[ki,ka,kj]

                # Remove zero/padded elements from denominator
                eia = LARGE_DENOM * np.ones((nocc, nvir), dtype=mo_energy[0].dtype)
                n0_ovp_ia = np.ix_(nonzero_opadding[ki], nonzero_vpadding[ka])
                eia[n0_ovp_ia] = (mo_e_o[ki][:,None] - mo_e_v[ka])[n0_ovp_ia]

                ejb = LARGE_DENOM * np.ones((nocc, nvir), dtype=mo_energy[0].dtype)
                n0_ovp_jb = np.ix_(nonzero_opadding[kj], nonzero_vpadding[kb])
                ejb[n0_ovp_jb] = (mo_e_o[kj][:,None] - mo_e_v[kb])[n0_ovp_jb]

                eijab = lib.direct_sum('ia,jb->ijab',eia,ejb)
                t2_ijab = np.conj(oovv_ij[ka]/eijab)
                if with_t2:
                    t2[ki, kj, ka] = t2_ijab
                edi = einsum('ijab,ijab', t2_ijab, oovv_ij[ka]).real * 2
                exi = -einsum('ijab,ijba', t2_ijab, oovv_ij[kb]).real
                emp2_ss += edi*0.5 + exi
                emp2_os += edi*0.5

    log.timer("KMP2", *cput0)

    emp2_ss /= nkpts
    emp2_os /= nkpts
    emp2 = lib.tag_array(emp2_ss+emp2_os, e_corr_ss=emp2_ss, e_corr_os=emp2_os)

    return emp2, t2


def _init_mp_df_eris(mp):
    """Compute 3-center electron repulsion integrals, i.e. (L|ov),
    where `L` denotes DF auxiliary basis functions and `o` and `v` occupied and virtual
    canonical crystalline orbitals. Note that `o` and `v` contain kpt indices `ko` and `kv`,
    and the third kpt index `kL` is determined by the conservation of momentum.

    Arguments:
        mp (KMP2) -- A KMP2 instance

    Returns:
        Lov (numpy.ndarray) -- 3-center DF ints, with shape (nkpts, nkpts, naux, nocc, nvir)
    """
    from pyscf.ao2mo import _ao2mo
    from pyscf.pbc.lib.kpts_helper import gamma_point

    log = logger.Logger(mp.stdout, mp.verbose)

    if mp._scf.with_df._cderi is None:
        mp._scf.with_df.build()

    cell = mp._scf.cell
    if cell.dimension == 2:
        # 2D ERIs are not positive definite. The 3-index tensors are stored in
        # two part. One corresponds to the positive part and one corresponds
        # to the negative part. The negative part is not considered in the
        # DF-driven CCSD implementation.
        raise NotImplementedError

    nocc = mp.nocc
    nmo = mp.nmo
    nvir = nmo - nocc
    nao = cell.nao_nr()

    mo_coeff = _add_padding(mp, mp.mo_coeff, mp.mo_energy)[0]
    kpts = mp.kpts
    nkpts = len(kpts)
    if gamma_point(kpts):
        dtype = np.double
    else:
        dtype = np.complex128
    dtype = np.result_type(dtype, *mo_coeff)
    Lov = np.empty((nkpts, nkpts), dtype=object)

    cput0 = (logger.process_clock(), logger.perf_counter())

    bra_start = 0
    bra_end = nocc
    ket_start = nmo+nocc
    ket_end = ket_start + nvir
    with df.CDERIArray(mp._scf.with_df._cderi) as cderi_array:
        tao = []
        ao_loc = None
        for ki in range(nkpts):
            for kj in range(nkpts):
                Lpq_ao = cderi_array[ki,kj]

                mo = np.hstack((mo_coeff[ki], mo_coeff[kj]))
                mo = np.asarray(mo, dtype=dtype, order='F')
                if dtype == np.double:
                    out = _ao2mo.nr_e2(Lpq_ao, mo, (bra_start, bra_end, ket_start, ket_end), aosym='s2')
                else:
                    #Note: Lpq.shape[0] != naux if linear dependency is found in auxbasis
                    if Lpq_ao[0].size != nao**2:  # aosym = 's2'
                        Lpq_ao = lib.unpack_tril(Lpq_ao).astype(np.complex128)
                    out = _ao2mo.r_e2(Lpq_ao, mo, (bra_start, bra_end, ket_start, ket_end), tao, ao_loc)
                Lov[ki, kj] = out.reshape(-1, nocc, nvir)

    log.timer_debug1("transforming DF-MP2 integrals", *cput0)

    return Lov


def _padding_k_idx(nmo, nocc, kind="split"):
    """A convention used for padding vectors, matrices and tensors in case when occupation numbers depend on the
    k-point index.
    Args:
        nmo (Iterable): k-dependent orbital number;
        nocc (Iterable): k-dependent occupation numbers;
        kind (str): either "split" (occupied and virtual spaces are split) or "joint" (occupied and virtual spaces are
        the joint;

    Returns:
        Two lists corresponding to the occupied and virtual spaces for kind="split". Each list contains integer arrays
        with indexes pointing to actual non-zero entries in the padded vector/matrix/tensor. If kind="joint", a single
        list of arrays is returned corresponding to the entire MO space.
    """
    if kind not in ("split", "joint"):
        raise ValueError("The 'kind' argument must be one of 'split', 'joint'")

    if kind == "split":
        indexes_o = []
        indexes_v = []
    else:
        indexes = []

    nocc = np.array(nocc)
    nmo = np.array(nmo)
    nvirt = nmo - nocc
    dense_o = np.amax(nocc)
    dense_v = np.amax(nvirt)
    dense_nmo = dense_o + dense_v

    for k_o, k_nmo in zip(nocc, nmo):
        k_v = k_nmo - k_o
        if kind == "split":
            indexes_o.append(np.arange(k_o))
            indexes_v.append(np.arange(dense_v - k_v, dense_v))
        else:
            indexes.append(np.concatenate((
                np.arange(k_o),
                np.arange(dense_nmo - k_v, dense_nmo),
            )))

    if kind == "split":
        return indexes_o, indexes_v

    else:
        return indexes


def padding_k_idx(mp, kind="split"):
    """A convention used for padding vectors, matrices and tensors in case when occupation numbers depend on the
    k-point index.

    This implementation stores k-dependent Fock and other matrix in dense arrays with additional dimensions
    corresponding to k-point indexes. In case when the occupation numbers depend on the k-point index (i.e. a metal) or
    when some k-points have more Bloch basis functions than others the corresponding data structure has to be padded
    with entries that are not used (fictitious occupied and virtual degrees of freedom). Current convention stores these
    states at the Fermi level as shown in the following example.

    +----+--------+--------+--------+
    |    |  k=0   |  k=1   |  k=2   |
    |    +--------+--------+--------+
    |    | nocc=2 | nocc=3 | nocc=2 |
    |    | nvir=4 | nvir=3 | nvir=3 |
    +====+========+========+========+
    | v3 |  k0v3  |  k1v2  |  k2v2  |
    +----+--------+--------+--------+
    | v2 |  k0v2  |  k1v1  |  k2v1  |
    +----+--------+--------+--------+
    | v1 |  k0v1  |  k1v0  |  k2v0  |
    +----+--------+--------+--------+
    | v0 |  k0v0  |        |        |
    +====+========+========+========+
    |          Fermi level          |
    +====+========+========+========+
    | o2 |        |  k1o2  |        |
    +----+--------+--------+--------+
    | o1 |  k0o1  |  k1o1  |  k2o1  |
    +----+--------+--------+--------+
    | o0 |  k0o0  |  k1o0  |  k2o0  |
    +----+--------+--------+--------+

    In the above example, `get_nmo(mp, per_kpoint=True) == (6, 6, 5)`, `get_nocc(mp, per_kpoint) == (2, 3, 2)`. The
    resulting dense `get_nmo(mp) == 7` and `get_nocc(mp) == 3` correspond to padded dimensions. This function will
    return the following indexes corresponding to the filled entries of the above table:

    >>> padding_k_idx(mp, kind="split")
    ([(0, 1), (0, 1, 2), (0, 1)], [(0, 1, 2, 3), (1, 2, 3), (1, 2, 3)])

    >>> padding_k_idx(mp, kind="joint")
    [(0, 1, 3, 4, 5, 6), (0, 1, 2, 4, 5, 6), (0, 1, 4, 5, 6)]

    Args:
        mp (:class:`MP2`): An instantiation of an SCF or post-Hartree-Fock object.
        kind (str): either "split" (occupied and virtual spaces are split) or "joint" (occupied and virtual spaces are
        the joint;

    Returns:
        Two lists corresponding to the occupied and virtual spaces for kind="split". Each list contains integer arrays
        with indexes pointing to actual non-zero entries in the padded vector/matrix/tensor. If kind="joint", a single
        list of arrays is returned corresponding to the entire MO space.
    """
    return _padding_k_idx(mp.get_nmo(per_kpoint=True), mp.get_nocc(per_kpoint=True), kind=kind)


def padded_mo_energy(mp, mo_energy):
    """
    Pads energies of active MOs.

    Args:
        mp (:class:`MP2`): An instantiation of an SCF or post-Hartree-Fock object.
        mo_energy (ndarray): original non-padded molecular energies;

    Returns:
        Padded molecular energies.
    """
    frozen_mask = get_frozen_mask(mp)
    padding_convention = padding_k_idx(mp, kind="joint")
    nkpts = mp.nkpts

    result = np.zeros((nkpts, mp.nmo), dtype=mo_energy[0].dtype)
    for k in range(nkpts):
        result[np.ix_([k], padding_convention[k])] = mo_energy[k][frozen_mask[k]]

    return result


def padded_mo_coeff(mp, mo_coeff):
    """
    Pads coefficients of active MOs.

    Args:
        mp (:class:`MP2`): An instantiation of an SCF or post-Hartree-Fock object.
        mo_coeff (ndarray): original non-padded molecular coefficients;

    Returns:
        Padded molecular coefficients.
    """
    frozen_mask = get_frozen_mask(mp)
    padding_convention = padding_k_idx(mp, kind="joint")
    nkpts = mp.nkpts

    result = np.zeros((nkpts, mo_coeff[0].shape[0], mp.nmo), dtype=mo_coeff[0].dtype)
    for k in range(nkpts):
        result[np.ix_([k], np.arange(result.shape[1]), padding_convention[k])] = mo_coeff[k][:, frozen_mask[k]]

    return result


def _frozen_sanity_check(frozen, mo_occ, kpt_idx):
    '''Performs a few sanity checks on the frozen array and mo_occ.

    Specific tests include checking for duplicates within the frozen array.

    Args:
        frozen (array_like of int): The orbital indices that will be frozen.
        mo_occ (:obj:`ndarray` of int): The occupation number for each orbital
            resulting from a mean-field-like calculation.
        kpt_idx (int): The k-point that `mo_occ` and `frozen` belong to.

    '''
    frozen = np.array(frozen)
    nocc = np.count_nonzero(mo_occ > 0)

    assert nocc, 'No occupied orbitals?\n\nnocc = %s\nmo_occ = %s' % (nocc, mo_occ)
    all_frozen_unique = (len(frozen) - len(np.unique(frozen))) == 0
    if not all_frozen_unique:
        raise RuntimeError('Frozen orbital list contains duplicates!\n\nkpt_idx %s\n'
                           'frozen %s' % (kpt_idx, frozen))
    if len(frozen) > 0 and np.max(frozen) > len(mo_occ) - 1:
        raise RuntimeError('Freezing orbital not in MO list!\n\nkpt_idx %s\n'
                           'frozen %s\nmax orbital idx %s' % (kpt_idx, frozen, len(mo_occ) - 1))


def get_nocc(mp, per_kpoint=False):
    '''Number of occupied orbitals for k-point calculations.

    Number of occupied orbitals for use in a calculation with k-points, taking into
    account frozen orbitals.

    Args:
        mp (:class:`MP2`): An instantiation of an SCF or post-Hartree-Fock object.
        per_kpoint (bool, optional): True returns the number of occupied
            orbitals at each k-point.  False gives the max of this list.

    Returns:
        nocc (int, list of int): Number of occupied orbitals. For return type, see description of arg
            `per_kpoint`.

    '''
    for i, moocc in enumerate(mp.mo_occ):
        if np.any(moocc % 1 != 0):
            raise RuntimeError("Fractional occupation numbers encountered @ kp={:d}: {}. This may have been caused by "
                               "smearing of occupation numbers in the mean-field calculation. If so, consider "
                               "executing mf.smearing_method = False; mf.mo_occ = mf.get_occ() prior to calling "
                               "this".format(i, moocc))
    if mp._nocc is not None:
        return mp._nocc
    elif mp.frozen is None:
        nocc = [np.count_nonzero(mp.mo_occ[ikpt]) for ikpt in range(mp.nkpts)]
    elif isinstance(mp.frozen, (int, np.integer)):
        nocc = [(np.count_nonzero(mp.mo_occ[ikpt]) - mp.frozen) for ikpt in range(mp.nkpts)]
    elif isinstance(mp.frozen[0], (int, np.integer)):
        [_frozen_sanity_check(mp.frozen, mp.mo_occ[ikpt], ikpt) for ikpt in range(mp.nkpts)]
        nocc = []
        for ikpt in range(mp.nkpts):
            max_occ_idx = np.max(np.where(mp.mo_occ[ikpt] > 0))
            frozen_nocc = np.sum(np.array(mp.frozen) <= max_occ_idx)
            nocc.append(np.count_nonzero(mp.mo_occ[ikpt]) - frozen_nocc)
    elif isinstance(mp.frozen[0], (list, np.ndarray)):
        nkpts = len(mp.frozen)
        if nkpts != mp.nkpts:
            raise RuntimeError('Frozen list has a different number of k-points (length) than passed in mean-field/'
                               'correlated calculation.  \n\nCalculation nkpts = %d, frozen list = %s '
                               '(length = %d)' % (mp.nkpts, mp.frozen, nkpts))
        [_frozen_sanity_check(frozen, mo_occ, ikpt) for ikpt, frozen, mo_occ in zip(range(nkpts), mp.frozen, mp.mo_occ)]

        nocc = []
        for ikpt, frozen in enumerate(mp.frozen):
            max_occ_idx = np.max(np.where(mp.mo_occ[ikpt] > 0))
            frozen_nocc = np.sum(np.array(frozen) <= max_occ_idx)
            nocc.append(np.count_nonzero(mp.mo_occ[ikpt]) - frozen_nocc)
    else:
        raise NotImplementedError

    assert any(np.array(nocc) > 0), ('Must have occupied orbitals! \n\nnocc %s\nfrozen %s\nmo_occ %s' %
           (nocc, mp.frozen, mp.mo_occ))

    if not per_kpoint:
        nocc = np.amax(nocc)

    return nocc


def get_nmo(mp, per_kpoint=False):
    '''Number of orbitals for k-point calculations.

    Number of orbitals for use in a calculation with k-points, taking into account
    frozen orbitals.

    Note:
        If `per_kpoint` is False, then the number of orbitals here is equal to max(nocc) + max(nvir),
        where each max is done over all k-points.  Otherwise the number of orbitals is returned
        as a list of number of orbitals at each k-point.

    Args:
        mp (:class:`MP2`): An instantiation of an SCF or post-Hartree-Fock object.
        per_kpoint (bool, optional): True returns the number of orbitals at each k-point.
            For a description of False, see Note.

    Returns:
        nmo (int, list of int): Number of orbitals. For return type, see description of arg
            `per_kpoint`.

    '''
    if mp._nmo is not None:
        return mp._nmo

    if mp.frozen is None:
        nmo = [len(mp.mo_occ[ikpt]) for ikpt in range(mp.nkpts)]
    elif isinstance(mp.frozen, (int, np.integer)):
        nmo = [len(mp.mo_occ[ikpt]) - mp.frozen for ikpt in range(mp.nkpts)]
    elif isinstance(mp.frozen[0], (int, np.integer)):
        [_frozen_sanity_check(mp.frozen, mp.mo_occ[ikpt], ikpt) for ikpt in range(mp.nkpts)]
        nmo = [len(mp.mo_occ[ikpt]) - len(mp.frozen) for ikpt in range(mp.nkpts)]
    elif isinstance(mp.frozen, (list, np.ndarray)):
        nkpts = len(mp.frozen)
        if nkpts != mp.nkpts:
            raise RuntimeError('Frozen list has a different number of k-points (length) than passed in mean-field/'
                               'correlated calculation.  \n\nCalculation nkpts = %d, frozen list = %s '
                               '(length = %d)' % (mp.nkpts, mp.frozen, nkpts))
        [_frozen_sanity_check(fro, mo_occ, ikpt) for ikpt, fro, mo_occ in zip(range(nkpts), mp.frozen, mp.mo_occ)]

        nmo = [len(mp.mo_occ[ikpt]) - len(mp.frozen[ikpt]) for ikpt in range(nkpts)]
    else:
        raise NotImplementedError

    assert all(np.array(nmo) > 0), ('Must have a positive number of orbitals!\n\nnmo %s\nfrozen %s\nmo_occ %s' %
           (nmo, mp.frozen, mp.mo_occ))

    if not per_kpoint:
        # Depending on whether there are more occupied bands, we want to make sure that
        # nmo has enough room for max(nocc) + max(nvir) number of orbitals for occupied
        # and virtual space
        nocc = mp.get_nocc(per_kpoint=True)
        nmo = np.max(nocc) + np.max(np.array(nmo) - np.array(nocc))

    return nmo


def get_frozen_mask(mp):
    '''Boolean mask for orbitals in k-point post-HF method.

    Creates a boolean mask to remove frozen orbitals and keep other orbitals for post-HF
    calculations.

    Args:
        mp (:class:`MP2`): An instantiation of an SCF or post-Hartree-Fock object.

    Returns:
        moidx (list of :obj:`ndarray` of `bool`): Boolean mask of orbitals to include.

    '''
    moidx = [np.ones(x.size, dtype=bool) for x in mp.mo_occ]
    if mp.frozen is None:
        pass
    elif isinstance(mp.frozen, (int, np.integer)):
        for idx in moidx:
            idx[:mp.frozen] = False
    elif isinstance(mp.frozen[0], (int, np.integer)):
        frozen = list(mp.frozen)
        for idx in moidx:
            idx[frozen] = False
    elif isinstance(mp.frozen[0], (list, np.ndarray)):
        nkpts = len(mp.frozen)
        if nkpts != mp.nkpts:
            raise RuntimeError('Frozen list has a different number of k-points (length) than passed in mean-field/'
                               'correlated calculation.  \n\nCalculation nkpts = %d, frozen list = %s '
                               '(length = %d)' % (mp.nkpts, mp.frozen, nkpts))
        [_frozen_sanity_check(fro, mo_occ, ikpt) for ikpt, fro, mo_occ in zip(range(nkpts), mp.frozen, mp.mo_occ)]
        for ikpt, kpt_occ in enumerate(moidx):
            kpt_occ[mp.frozen[ikpt]] = False
    else:
        raise NotImplementedError

    return moidx


def _add_padding(mp, mo_coeff, mo_energy):
    nmo = mp.nmo

    # Check if these are padded mo coefficients and energies and/or if some orbitals are frozen.
    if (mp.frozen is not None) or (not np.all([x.shape[1] == nmo for x in mo_coeff])):
        mo_coeff = padded_mo_coeff(mp, mo_coeff)

    if (mp.frozen is not None) or (not np.all([x.shape[0] == nmo for x in mo_energy])):
        mo_energy = padded_mo_energy(mp, mo_energy)
    return mo_coeff, mo_energy


def make_rdm1(mp, t2=None, kind="compact"):
    r"""
    Spin-traced one-particle density matrix in the MO basis representation.
    The occupied-virtual orbital response is not included.

    dm1[p,q] = <q_alpha^\dagger p_alpha> + <q_beta^\dagger p_beta>

    The convention of 1-pdm is based on McWeeney's book, Eq (5.4.20).
    The contraction between 1-particle Hamiltonian and rdm1 is
    E = einsum('pq,qp', h1, rdm1)

    Args:
        mp (KMP2): a KMP2 kernel object;
        t2 (ndarray): a t2 MP2 tensor;
        kind (str): either 'compact' or 'padded' - defines behavior for k-dependent MO basis sizes;

    Returns:
        A k-dependent single-particle density matrix.
    """
    if kind not in ("compact", "padded"):
        raise ValueError("The 'kind' argument should be either 'compact' or 'padded'")
    d_imds = _gamma1_intermediates(mp, t2=t2)
    result = []
    padding_idxs = padding_k_idx(mp, kind="joint")
    for (oo, vv), idxs in zip(zip(*d_imds), padding_idxs):
        oo += np.eye(*oo.shape)
        d = block_diag(oo, vv)
        d += d.conj().T
        if kind == "padded":
            result.append(d)
        else:
            result.append(d[np.ix_(idxs, idxs)])
    return result


def make_rdm2(mp, t2=None, kind="compact"):
    r'''
    Spin-traced two-particle density matrix in MO basis

    .. math::

        dm2[p,q,r,s] = \sum_{\sigma,\tau} <p_\sigma^\dagger r_\tau^\dagger s_\tau q_\sigma>

    Note the contraction between ERIs (in Chemist's notation) and rdm2 is
    E = einsum('pqrs,pqrs', eri, rdm2)
    '''
    if kind not in ("compact", "padded"):
        raise ValueError("The 'kind' argument should be either 'compact' or 'padded'")
    if t2 is None: t2 = mp.t2
    dm1 = mp.make_rdm1(t2, "padded")
    nmo = mp.nmo
    nocc = mp.nocc
    nkpts = mp.nkpts
    dtype = t2.dtype

    dm2 = np.zeros((nkpts,nkpts,nkpts,nmo,nmo,nmo,nmo),dtype=dtype)
    for ki in range(nkpts):
        for kj in range(nkpts):
            for ka in range(nkpts):
                kb = mp.khelper.kconserv[ki, ka, kj]
                dovov = t2[ki, kj, ka].transpose(0,2,1,3) * 2 - t2[kj, ki, ka].transpose(1,2,0,3)
                dovov *= 2
                dm2[ki,ka,kj,:nocc,nocc:,:nocc,nocc:] = dovov
                dm2[ka,ki,kb,nocc:,:nocc,nocc:,:nocc] = dovov.transpose(1,0,3,2).conj()

    occidx = padding_k_idx(mp, kind="split")[0]
    for ki in range(nkpts):
        for i in occidx[ki]:
            dm1[ki][i,i] -= 2

    for ki in range(nkpts):
        for kp in range(nkpts):
            for i in occidx[ki]:
                dm2[ki,ki,kp,i,i,:,:] += dm1[kp].T * 2
                dm2[kp,kp,ki,:,:,i,i] += dm1[kp].T * 2
                dm2[kp,ki,ki,:,i,i,:] -= dm1[kp].T
                dm2[ki,kp,kp,i,:,:,i] -= dm1[kp]

    for ki in range(nkpts):
        for kj in range(nkpts):
            for i in occidx[ki]:
                for j in occidx[kj]:
                    dm2[ki,ki,kj,i,i,j,j] += 4
                    dm2[ki,kj,kj,i,j,j,i] -= 2

    if kind == "padded":
        return dm2
    else:
        idx = padding_k_idx(mp, kind="joint")
        result = []
        for kp in range(nkpts):
            for kq in range(nkpts):
                for kr in range(nkpts):
                    ks = mp.khelper.kconserv[kp, kq, kr]
                    result.append(dm2[kp,kq,kr][np.ix_(idx[kp],idx[kq],idx[kr],idx[ks])])
        return result


def _gamma1_intermediates(mp, t2=None):
    # Memory optimization should be here
    if t2 is None:
        t2 = mp.t2
    if t2 is None:
        raise NotImplementedError("Run kmp2.kernel with `with_t2=True`")
    nmo = mp.nmo
    nocc = mp.nocc
    nvir = nmo - nocc
    nkpts = mp.nkpts
    dtype = t2.dtype

    dm1occ = np.zeros((nkpts, nocc, nocc), dtype=dtype)
    dm1vir = np.zeros((nkpts, nvir, nvir), dtype=dtype)

    for ki in range(nkpts):
        for kj in range(nkpts):
            for ka in range(nkpts):
                kb = mp.khelper.kconserv[ki, ka, kj]

                dm1vir[kb] += einsum('ijax,ijay->yx', t2[ki][kj][ka].conj(), t2[ki][kj][ka]) * 2 -\
                              einsum('ijax,ijya->yx', t2[ki][kj][ka].conj(), t2[ki][kj][kb])
                dm1occ[kj] += einsum('ixab,iyab->xy', t2[ki][kj][ka].conj(), t2[ki][kj][ka]) * 2 -\
                              einsum('ixab,iyba->xy', t2[ki][kj][ka].conj(), t2[ki][kj][kb])
    return -dm1occ, dm1vir


class KMP2(mp2.MP2):
    _keys = {'kpts', 'nkpts', 'khelper'}

    def __init__(self, mf, frozen=None, mo_coeff=None, mo_occ=None):

        if mo_coeff is None: mo_coeff = mf.mo_coeff
        if mo_occ is None: mo_occ = mf.mo_occ

        self.mol = mf.mol
        self._scf = mf
        self.verbose = self.mol.verbose
        self.stdout = self.mol.stdout
        self.max_memory = mf.max_memory

        self.frozen = frozen
        if isinstance(self._scf.with_df, df.GDF):
            self.with_df_ints = True
        else:
            self.with_df_ints = False

##################################################
# don't modify the following attributes, they are not input options
        self.kpts = mf.kpts
        if isinstance(self.kpts, libkpts.KPoints):
            self.nkpts = self.kpts.nkpts
            self.khelper = kpts_helper.KptsHelper(mf.cell, mf.kpts.kpts)
            #padding has to be after transformation
            self.mo_energy = self.kpts.transform_mo_energy(mf.mo_energy)
            self.mo_coeff = self.kpts.transform_mo_coeff(mo_coeff)
            self.mo_occ = self.kpts.transform_mo_occ(mo_occ)
        else:
            self.nkpts = len(self.kpts)
            self.khelper = kpts_helper.KptsHelper(mf.cell, mf.kpts)
            self.mo_energy = mf.mo_energy
            self.mo_coeff = mo_coeff
            self.mo_occ = mo_occ
        self._nocc = None
        self._nmo = None
        self.e_hf = None
        self.e_corr = None
        self.e_corr_ss = None
        self.e_corr_os = None
        self.t2 = None

    get_nocc = get_nocc
    get_nmo = get_nmo
    get_frozen_mask = get_frozen_mask
    make_rdm1 = make_rdm1
    make_rdm2 = make_rdm2

    def dump_flags(self):
        logger.info(self, "")
        logger.info(self, "******** %s ********", self.__class__)
        logger.info(self, "nkpts = %d", self.nkpts)
        logger.info(self, "nocc = %d", self.nocc)
        logger.info(self, "nmo = %d", self.nmo)
        logger.info(self, "with_df_ints = %s", self.with_df_ints)

        if self.frozen is not None:
            logger.info(self, "frozen orbitals = %s", self.frozen)
        logger.info(
            self,
            "max_memory %d MB (current use %d MB)",
            self.max_memory,
            lib.current_memory()[0],
        )
        return self

    def kernel(self, mo_energy=None, mo_coeff=None, with_t2=WITH_T2):
        if mo_energy is None:
            mo_energy = self.mo_energy
        if mo_coeff is None:
            mo_coeff = self.mo_coeff
        if mo_energy is None or mo_coeff is None:
            log = logger.Logger(self.stdout, self.verbose)
            log.warn('mo_coeff, mo_energy are not given.\n'
                     'You may need to call mf.kernel() to generate them.')
            raise RuntimeError

        self.e_hf = self.get_e_hf(mo_coeff=mo_coeff)

        mo_coeff, mo_energy = _add_padding(self, mo_coeff, mo_energy)

        self.e_corr, self.t2 = \
                kernel(self, mo_energy, mo_coeff, verbose=self.verbose, with_t2=with_t2)

        self.e_corr_ss = getattr(self.e_corr, 'e_corr_ss', 0)
        self.e_corr_os = getattr(self.e_corr, 'e_corr_os', 0)
        self.e_corr = float(self.e_corr)

        self._finalize()

        return self.e_corr, self.t2

    to_gpu = lib.to_gpu

KRMP2 = KMP2


from pyscf.pbc import scf
scf.khf.KRHF.MP2 = lib.class_as_method(KRMP2)
scf.kghf.KGHF.MP2 = None
scf.krohf.KROHF.MP2 = None


if __name__ == '__main__':
    from pyscf.pbc import gto, scf, mp

    cell = gto.Cell()
    cell.atom='''
    C 0.000000000000   0.000000000000   0.000000000000
    C 1.685068664391   1.685068664391   1.685068664391
    '''
    cell.basis = 'gth-szv'
    cell.pseudo = 'gth-pade'
    cell.a = '''
    0.000000000, 3.370137329, 3.370137329
    3.370137329, 0.000000000, 3.370137329
    3.370137329, 3.370137329, 0.000000000'''
    cell.unit = 'B'
    cell.verbose = 5
    cell.build()

    # Running HF and MP2 with 1x1x2 Monkhorst-Pack k-point mesh
    kmf = scf.KRHF(cell, kpts=cell.make_kpts([1,1,2]), exxdiv=None)
    ehf = kmf.kernel()

    mymp = mp.KMP2(kmf)
    emp2, t2 = mymp.kernel()
    print(emp2 - -0.204721432828996)



#!/usr/bin/env python
# Copyright 2024 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
import scipy.linalg
from pyscf.lib.parameters import MAX_MEMORY
from pyscf.lib import logger
from pyscf.lib.exceptions import LinearDependencyError
from pyscf.lib.linalg_helper import _sort_elast, _outprod_to_subspace

# Add at most 20 states in each iteration
MAX_SPACE_INC = 20

def eigh(aop, x0, precond, tol_residual=1e-5, lindep=1e-12, nroots=1,
         x0sym=None, pick=None, max_cycle=50, max_memory=MAX_MEMORY,
         verbose=logger.WARN):
    '''
    Solve symmetric eigenvalues.

    This solver is similar to the `linalg_helper.davidson` solver, with
    optimizations for performance and wavefunction symmetry, specifically
    tailored for linear response methods.

    Args:
        aop : function(x) => array_like_x
            The matrix-vector product operator.
        x0 : 1D array or a list of 1D array
            Initial guess.
        precond : function(dx, e) => array_like_dx
            Preconditioner to generate new trial vector. The argument dx is a
            residual vector ``A*x0-e*x0``; e is the eigenvalue.

    Kwargs:
        tol_residual : float
            Convergence tolerance for the norm of residual vector ``A*x0-e*x0``.
        lindep : float
            Linear dependency threshold.  The function is terminated when the
            smallest eigenvalue of the metric of the trial vectors is lower
            than this threshold.
        nroots : int
            Number of eigenvalues to be computed.
        x0sym:
            The symmetry label for each initial guess vectors.
        pick : function(w,v,nroots) => (e[idx], w[:,idx], idx)
            Function to filter eigenvalues and eigenvectors.
        max_cycle : int
            max number of iterations.
        max_memory : int or float
            Allowed memory in MB.

    Returns:
        e : list of floats
            Eigenvalues.
        c : list of 1D arrays
            Eigenvectors.
    '''

    assert callable(pick)
    assert callable(precond)

    log = logger.new_logger(verbose)

    if isinstance(x0, np.ndarray) and x0.ndim == 1:
        x0 = x0[None,:]
    x0 = np.asarray(x0)

    x0_size = x0.shape[1]
    if MAX_SPACE_INC is None:
        space_inc = nroots
    else:
        # Adding too many trial bases in each iteration may cause larger errors
        space_inc = max(nroots, min(MAX_SPACE_INC, x0_size//2))

    max_space = int(max_memory*1e6/8/x0_size / 2 - nroots - space_inc)
    if max_space < nroots * 4 < x0_size:
        log.warn('Not enough memory to store trial space in _lr_eig.eigh')
    max_space = max(max_space, nroots * 4)
    max_space = min(max_space, x0_size)
    log.debug(f'Set max_space {max_space}, space_inc {space_inc}')

    xs = np.zeros((0, x0_size))
    ax = np.zeros((0, x0_size))
    e = w = v = None
    conv_last = conv = np.zeros(nroots, dtype=bool)
    xt = x0

    if x0sym is not None:
        xt_ir = np.asarray(x0sym)
        xs_ir = np.array([], dtype=xt_ir.dtype)

    for icyc in range(max_cycle):
        xt, xt_idx = _qr(xt, lindep)
        # Generate at most space_inc trial vectors
        xt = xt[:space_inc]
        xt_idx = xt_idx[:space_inc]

        row0 = len(xs)
        axt = aop(xt)
        xs = np.vstack([xs, xt])
        ax = np.vstack([ax, axt])
        if x0sym is not None:
            xs_ir = np.hstack([xs_ir, xt_ir[xt_idx]])

        # Compute heff = xs.conj().dot(ax.T)
        if w is None:
            heff = xs.conj().dot(ax.T)
        else:
            hsub = xt.conj().dot(ax.T)
            heff = np.block([[np.diag(w), hsub[:,:row0].conj().T],
                             [hsub[:,:row0], hsub[:,row0:]]])

        if x0sym is None:
            w, v = scipy.linalg.eigh(heff)
        else:
            # Diagonalize within eash symmetry sectors
            row1 = len(xs)
            w = np.empty(row1)
            v = np.zeros((row1, row1))
            v_ir = []
            i1 = 0
            for ir in set(xs_ir):
                idx = np.where(xs_ir == ir)[0]
                i0, i1 = i1, i1 + idx.size
                w_sub, v_sub = scipy.linalg.eigh(heff[idx[:,None],idx])
                w[i0:i1] = w_sub
                v[idx,i0:i1] = v_sub
                v_ir.append([ir] * idx.size)
            w_idx = np.argsort(w)
            w = w[w_idx]
            v = v[:,w_idx]
            xs_ir = np.hstack(v_ir)[w_idx]

        w, v, idx = pick(w, v, nroots, locals())
        if x0sym is not None:
            xs_ir = xs_ir[idx]
        if len(w) == 0:
            raise RuntimeError('Not enough eigenvalues')

        e, elast = w[:nroots], e
        if elast is None:
            de = e
        elif elast.size != e.size:
            log.debug('Number of roots different from the previous step (%d,%d)',
                      e.size, elast.size)
            de = e
        else:
            # mapping to previous eigenvectors
            vlast = np.eye(nroots)
            elast, conv_last = _sort_elast(elast, conv, vlast,
                                           v[:nroots,:nroots], log)
            de = e - elast

        xs = v.T.dot(xs)
        ax = v.T.dot(ax)
        if len(xs) * 2 > max_space:
            row0 = max(nroots, max_space-space_inc)
            xs = xs[:row0]
            ax = ax[:row0]
            w = w[:row0]
            if x0sym is not None:
                xs_ir = xs_ir[:row0]

        t_size = max(nroots, max_space-len(xs))
        xt = -w[:t_size,None] * xs[:t_size]
        xt += ax[:t_size]
        if x0sym is not None:
            xt_ir = xs_ir[:t_size]

        dx_norm = np.linalg.norm(xt, axis=1)
        max_dx_norm = max(dx_norm[:nroots])
        conv = dx_norm[:nroots] < tol_residual
        for k, ek in enumerate(e[:nroots]):
            if conv[k] and not conv_last[k]:
                log.debug('root %d converged  |r|= %4.3g  e= %s  max|de|= %4.3g',
                          k, dx_norm[k], ek, de[k])
            else:
                log.debug1('root %d  |r|= %4.3g  e= %s  max|de|= %4.3g',
                          k, dx_norm[k], ek, de[k])
        ide = np.argmax(abs(de))
        if all(conv):
            log.debug('converged %d %d  |r|= %4.3g  e= %s  max|de|= %4.3g',
                      icyc, len(xs), max_dx_norm, e, de[ide])
            break

        # remove subspace linear dependency
        for k, xk in enumerate(xt):
            if dx_norm[k] > tol_residual:
                xt[k] = precond(xk, e[0])
        xt -= xs.conj().dot(xt.T).T.dot(xs)
        xt_norm = np.linalg.norm(xt, axis=1)

        remaining = []
        for k, xk in enumerate(xt):
            if dx_norm[k] > tol_residual and xt_norm[k]**2 > lindep:
                xt[k] /= xt_norm[k]
                remaining.append(k)
        if len(remaining) == 0:
            log.debug(f'Linear dependency in trial subspace. |r| for each state {dx_norm}')
            break

        xt = xt[remaining]
        log.debug1('Generate %d trial vectors. Drop %d vectors',
                   len(xt), dx_norm.size - len(xt))

        if x0sym is not None:
            xt_ir = xt_ir[remaining]
        norm_min = xt_norm[remaining].min()
        log.debug('davidson %d %d |r|= %4.3g  e= %s  max|de|= %4.3g  lindep= %4.3g',
                  icyc, len(xs), max_dx_norm, e, de[ide], norm_min)

    x0 = xs[:nroots]
    # Check whether the solver finds enough eigenvectors.
    if len(x0) < min(x0_size, nroots):
        log.warn(f'Not enough eigenvectors (len(x0)={len(x0)}, nroots={nroots})')

    return conv, e, x0

def eig(aop, x0, precond, tol_residual=1e-5, nroots=1, x0sym=None, pick=None,
        max_cycle=50, max_memory=MAX_MEMORY, lindep=1e-12, verbose=logger.WARN):
    '''
    Solver for linear response eigenvalues
    [ A    B] [X] = w [X]
    [-B* -A*] [Y]     [Y]

    subject to normalization X^2 - Y^2 = 1

    Reference:
      Olsen, Jensen, and Jorgenson, J Comput Phys, 74, 265,
      DOI: 10.1016/0021-9991(88)90081-2

    Args:
        aop : function(x) => array_like_x
            The matrix-vector product operator.
        x0 : 1D array or a list of 1D array
            Initial guess.
        precond : function(dx, e) => array_like_dx
            Preconditioner to generate new trial vector.
            The argument dx is a residual vector ``A*x0-e*x0``; e is the eigenvalue.

    Kwargs:
        tol_residual : float
            Convergence tolerance for the norm of residual vector ``A*x0-e*x0``.
        lindep : float
            Linear dependency threshold.
        nroots : int
            Number of eigenvalues to be computed.
        x0sym:
            The symmetry label for each initial guess vectors.
        pick : function(w,v,nroots) => (e[idx], w[:,idx], idx)
            Function to filter eigenvalues and eigenvectors.
        max_cycle : int
            max number of iterations.
        max_memory : int or float
            Allowed memory in MB.

    Returns:
        conv : list of booleans
            Whether each root is converged.
        e : list of floats
            Eigenvalues.
        c : list of 1D arrays
            Eigenvectors.
    '''

    assert callable(pick)
    assert callable(precond)

    log = logger.new_logger(verbose)

    if isinstance(x0, np.ndarray) and x0.ndim == 1:
        x0 = x0[None,:]
    x0 = np.asarray(x0)
    x0_size = x0.shape[1]
    if MAX_SPACE_INC is None:
        space_inc = nroots
    else:
        # Adding too many trial bases in each iteration may introduce more errors
        space_inc = max(nroots, min(MAX_SPACE_INC, x0_size//4))

    max_space = int(max_memory*1e6/8/(2*x0_size) / 2 - space_inc)
    if max_space < nroots * 4 < x0_size:
        log.warn('Not enough memory to store trial space in _lr_eig.eig')
        max_space = space_inc * 2
    max_space = max(max_space, nroots * 4)
    max_space = min(max_space, x0_size)
    log.debug(f'Set max_space {max_space}, space_inc {space_inc}')

    if x0sym is None:
        x0 = _symmetric_orth(x0)
    else:
        x0_ir = np.asarray(x0sym)
        x0_orth = []
        x0_orth_ir = []
        for ir in set(x0_ir):
            idx = np.where(x0_ir == ir)[0]
            xt_sub = _symmetric_orth(x0[idx])
            x0_orth.append(xt_sub)
            x0_orth_ir.append([ir] * len(xt_sub))
        if x0_orth:
            x0 = np.vstack(x0_orth)
            x0_ir = np.hstack(x0_orth_ir)
        else:
            x0 = []
        x0_orth = x0_orth_ir = xt_sub = None
    if len(x0) == 0:
        raise LinearDependencyError('Empty initial guess')

    heff = None
    e = None
    v = None
    vlast = None
    conv_last = conv = np.zeros(nroots, dtype=bool)

    half_size = x0[0].size // 2
    fresh_start = True
    for icyc in range(max_cycle):
        if fresh_start:
            vlast = None
            conv_last = conv = np.zeros(nroots, dtype=bool)
            xs = np.zeros((0, x0_size))
            ax = np.zeros((0, x0_size))
            row1 = 0
            xt = x0
            if x0sym is not None:
                xs_ir = x0_ir

        axt = aop(xt)
        xs = np.vstack([xs, xt])
        ax = np.vstack([ax, axt])
        row0, row1 = row1, row1+len(xt)

        if heff is None:
            dtype = np.result_type(axt, xt)
            heff = np.empty((max_space*2,max_space*2), dtype=dtype)

        h11 = xs[:row0].conj().dot(axt.T).astype(dtype)
        h21 = _conj_dot(xs[:row0], axt).astype(dtype)
        heff[0:row0*2:2, row0*2+0:row1*2:2] = h11
        heff[1:row0*2:2, row0*2+0:row1*2:2] = h21
        heff[0:row0*2:2, row0*2+1:row1*2:2] = -h21.conj()
        heff[1:row0*2:2, row0*2+1:row1*2:2] = -h11.conj()

        h11 = xt.conj().dot(ax.T).astype(dtype)
        h21 = _conj_dot(xt, ax).astype(dtype)
        heff[row0*2+0:row1*2:2, 0:row1*2:2] = h11
        heff[row0*2+1:row1*2:2, 0:row1*2:2] = h21
        heff[row0*2+0:row1*2:2, 1:row1*2:2] = -h21.conj()
        heff[row0*2+1:row1*2:2, 1:row1*2:2] = -h11.conj()

        if x0sym is None:
            w, v = scipy.linalg.eig(heff[:row1*2,:row1*2])
        else:
            # Diagonalize within eash symmetry sectors
            xs_ir2 = np.repeat(xs_ir, 2)
            w = np.empty(row1*2, dtype=np.complex128)
            v = np.zeros((row1*2, row1*2), dtype=np.complex128)
            v_ir = []
            i1 = 0
            for ir in set(xs_ir):
                idx = np.where(xs_ir2 == ir)[0]
                i0, i1 = i1, i1 + idx.size
                w_sub, v_sub = scipy.linalg.eig(heff[idx[:,None],idx])
                w[i0:i1] = w_sub
                v[idx,i0:i1] = v_sub
                v_ir.append([ir] * idx.size)
            v_ir = np.hstack(v_ir)

        w, v, idx = pick(w, v, nroots, locals())
        if x0sym is not None:
            v_ir = v_ir[idx]
        if len(w) == 0:
            raise RuntimeError('Not enough eigenvalues')

        w, e, elast = w[:space_inc], w[:nroots], e
        v = v[:,:space_inc]
        if vlast is not None:
            elast, conv_last = _sort_elast(elast, conv, vlast, v[:,:nroots], log)
        vlast = v[:,:nroots]

        if elast is None:
            de = e
        elif elast.size != e.size:
            log.debug('Number of roots different from the previous step (%d,%d)',
                      e.size, elast.size)
            de = e
        else:
            de = e - elast

        x0 = _gen_x0(v, xs)
        ax0 = xt = _gen_ax0(v, ax)
        xt -= w[:,None] * x0
        ax0 = None
        if x0sym is not None:
            xt_ir = v_ir[:space_inc]
            x0_ir = v_ir[:nroots]

        dx_norm = np.linalg.norm(xt, axis=1)
        max_dx_norm = max(dx_norm[:nroots])
        conv = dx_norm[:nroots] < tol_residual
        for k, ek in enumerate(e[:nroots]):
            if conv[k] and not conv_last[k]:
                log.debug('root %d converged  |r|= %4.3g  e= %s  max|de|= %4.3g',
                          k, dx_norm[k], ek, de[k])
            else:
                log.debug1('root %d  |r|= %4.3g  e= %s  max|de|= %4.3g',
                          k, dx_norm[k], ek, de[k])
        ide = np.argmax(abs(de))
        if all(conv):
            log.debug('converged %d %d  |r|= %4.3g  e= %s  max|de|= %4.3g',
                      icyc, len(xs), max_dx_norm, e, de[ide])
            break

        # remove subspace linear dependency
        for k, xk in enumerate(xt):
            if dx_norm[k] > tol_residual:
                xt[k] = precond(xk, e[0])

        xt -= xs.conj().dot(xt.T).T.dot(xs)
        c = _conj_dot(xs, xt)
        xt[:,:half_size] -= c.T.dot(xs[:,half_size:].conj())
        xt[:,half_size:] -= c.T.dot(xs[:,:half_size].conj())

        # Remove quasi linearly dependent bases, as they cause more numerical
        # errors in _symmetric_orth
        xt_norm = np.linalg.norm(xt, axis=1)
        xt_to_keep = (dx_norm > tol_residual) & (xt_norm > max(lindep**.5, tol_residual))
        xt = xt[xt_to_keep]
        if len(xt) > 0:
            xt /= xt_norm[xt_to_keep, None]
            if x0sym is None:
                xt = _symmetric_orth(xt)
            else:
                xt_ir = xt_ir[xt_to_keep]
                xt_orth = []
                xt_orth_ir = []
                for ir in set(xt_ir):
                    idx = np.where(xt_ir == ir)[0]
                    xt_sub = _symmetric_orth(xt[idx])
                    xt_orth.append(xt_sub)
                    xt_orth_ir.append([ir] * len(xt_sub))
                if xt_orth:
                    xt = np.vstack(xt_orth)
                    xs_ir = np.hstack([xs_ir, *xt_orth_ir])
                else:
                    xt = []
                xt_orth = xt_orth_ir = xt_sub = None

        if len(xt) == 0:
            log.debug(f'Linear dependency in trial subspace. |r| for each state {dx_norm}')
            break
        log.debug1('Generate %d trial vectors. Drop %d vectors',
                   len(xt), dx_norm.size - len(xt))

        xt_norm = np.linalg.norm(xt, axis=1)
        norm_min = xt_norm.min()
        log.debug('lr_eig %d %d  |r|= %4.3g  e= %s  max|de|= %4.3g  lindep= %4.3g',
                  icyc, len(xs), max_dx_norm, e, de[ide], norm_min)

        fresh_start = len(xs)+space_inc > max_space

    # Check whether the solver finds enough eigenvectors.
    h_dim = x0[0].size
    if len(x0) < min(h_dim, nroots):
        log.warn(f'Not enough eigenvectors (len(x0)={len(x0)}, nroots={nroots})')

    return conv[:nroots], e[:nroots], x0[:nroots]

def real_eig(aop, x0, precond, tol_residual=1e-5, nroots=1, x0sym=None, pick=None,
             max_cycle=50, max_memory=MAX_MEMORY, lindep=1e-12, verbose=logger.WARN):
    '''
    Solve linear response eigenvalues for real A and B matrices
    [ A  B] [X] = w [X]
    [-B -A] [Y]     [Y]

    subject to normalization X^2 - Y^2 = 1 . This function is based on the
    algorithm implemented in https://github.com/John-zzh/improved-Davidson-Algorithm

    Args:
        aop : function(x) => array_like_x
            The matrix-vector product operator.
        x0 : 1D array or a list of 1D array
            Initial guess.
        precond : function(dx, e) => array_like_dx
            Preconditioner to generate new trial vector.

    Kwargs:
        tol_residual : float
            Convergence tolerance for the norm of residual vector ``A*x0-e*x0``.
        lindep : float
            Linear dependency threshold.
        nroots : int
            Number of eigenvalues to be computed.
        x0sym:
            The symmetry label for each initial guess vectors.
        pick : function(w,v,nroots) => (e[idx], w[:,idx], idx)
            Function to filter eigenvalues and eigenvectors.
        max_cycle : int
            max number of iterations.
        max_memory : int or float
            Allowed memory in MB.

    Returns:
        conv : list of booleans
            Whether each root is converged.
        e : list of floats
            Eigenvalues.
        c : list of 1D arrays
            Eigenvectors.
    '''

    assert pick is None
    assert callable(precond)

    log = logger.new_logger(verbose)

    assert x0.ndim == 2
    A_size = x0.shape[1] // 2
    V = x0[:,:A_size]
    W = x0[:,A_size:]
    x0 = (V, W)
    if MAX_SPACE_INC is None:
        space_inc = nroots
    else:
        # Adding too many trial bases in each iteration may cause larger errors
        space_inc = max(nroots, min(MAX_SPACE_INC, A_size//2))

    max_space = int(max_memory*1e6/8/(4*A_size) / 2 - space_inc)
    if max_space < nroots * 4 < A_size:
        log.warn('Not enough memory to store trial space in _lr_eig.eig')
        max_space = space_inc * 2
    max_space = max(max_space, nroots * 4)
    max_space = min(max_space, A_size)
    log.debug(f'Set max_space {max_space}, space_inc {space_inc}')

    if x0sym is not None:
        x0_ir = np.asarray(x0sym)

    '''U1 = AV + BW
       U2 = AW + BV'''
    V_holder = np.empty((A_size, max_space), order='F')
    W_holder = np.empty_like(V_holder)
    U1_holder = np.empty_like(V_holder)
    U2_holder = np.empty_like(V_holder)

    a = np.empty((max_space*2,max_space*2))
    b = np.empty_like(a)
    sigma = np.empty_like(a)
    pi = np.empty_like(a)
    e = None
    v_sub = None
    vlast = None
    conv_last = conv = np.zeros(nroots, dtype=bool)

    fresh_start = True
    for icyc in range(max_cycle):
        if fresh_start:
            m0 = m1 = 0
            V, W = x0
            if x0sym is not None:
                xs_ir = xt_ir = x0_ir

        axt = aop(np.hstack([V, W]))
        U1 =  axt[:,:A_size]
        U2 = -axt[:,A_size:]
        m0, m1 = m1, m1+len(U1)
        V_holder [:,m0:m1] = V.T
        W_holder [:,m0:m1] = W.T
        U1_holder[:,m0:m1] = U1.T
        U2_holder[:,m0:m1] = U2.T

        '''
        a = np.dot(V.T, U1)
        a += np.dot(W.T, U2)
        b = np.dot(V.T, U2)
        b += np.dot(W.T, U1)
        sigma = np.dot(V.T, V)
        sigma -= np.dot(W.T, W)
        pi = np.dot(V.T, W)
        pi -= np.dot(W.T, V)
        a = (a + a.T) / 2
        b = (b + b.T) / 2
        sigma = (sigma + sigma.T) / 2
        pi = (pi - pi.T) / 2
        '''
        a_block  = _sym_dot(V_holder, U1_holder, m0, m1)
        a_block += _sym_dot(W_holder, U2_holder, m0, m1)
        b_block  = _sym_dot(V_holder, U2_holder, m0, m1)
        b_block += _sym_dot(W_holder, U1_holder, m0, m1)
        sigma_block  = _sym_dot(V_holder, V_holder, m0, m1)
        sigma_block -= _sym_dot(W_holder, W_holder, m0, m1)
        pi_block  = _asym_dot(V_holder, W_holder, m0, m1)
        pi_block -= _asym_dot(W_holder, V_holder, m0, m1)
        a[:m1,m0:m1] = a_block.T
        a[m0:m1,:m1] = a_block
        b[:m1,m0:m1] = b_block.T
        b[m0:m1,:m1] = b_block
        sigma[:m1,m0:m1] = sigma_block.T
        sigma[m0:m1,:m1] = sigma_block
        pi[:m1,m0:m1] = -pi_block.T
        pi[m0:m1,:m1] = pi_block

        if x0sym is None:
            omega, x, y = TDDFT_subspace_eigen_solver(
                a[:m1,:m1], b[:m1,:m1], sigma[:m1,:m1], pi[:m1,:m1], space_inc)
        else:
            # Diagonalize within eash symmetry sectors
            omega = np.empty(m1)
            x = np.zeros((m1, m1))
            y = np.zeros_like(x)
            v_ir = []
            i1 = 0
            for ir in set(xs_ir):
                idx = np.nonzero(xs_ir[:m1] == ir)[0]
                _w, _x, _y = TDDFT_subspace_eigen_solver(
                    a[idx[:,None],idx], b[idx[:,None],idx],
                    sigma[idx[:,None],idx], pi[idx[:,None],idx], idx.size)
                i0, i1 = i1, i1 + idx.size
                omega[i0:i1] = _w
                x[idx,i0:i1] = _x
                y[idx,i0:i1] = _y
                v_ir.append([ir] * _w.size)
            idx = np.argsort(omega)
            omega = omega[idx]
            v_ir = np.hstack(v_ir)[idx]
            x = x[:,idx]
            y = y[:,idx]

        w, e, elast = omega[:space_inc], omega[:nroots], e
        v_sub = x[:,:space_inc]
        if not fresh_start:
            elast, conv_last = _sort_elast(elast, conv, vlast, v_sub[:,:nroots], log)
        vlast = v_sub[:,:nroots]

        if elast is None:
            de = e
        elif elast.size != e.size:
            log.debug('Number of roots different from the previous step (%d,%d)',
                      e.size, elast.size)
            de = e
        else:
            de = e - elast

        x = x[:,:space_inc]
        y = y[:,:space_inc]
        X_full  = V_holder[:,:m1].dot(x)
        X_full += W_holder[:,:m1].dot(y)
        Y_full  = W_holder[:,:m1].dot(x)
        Y_full += V_holder[:,:m1].dot(y)
        x0 = (X_full[:,:nroots].T, Y_full[:,:nroots].T)
        # residuals
        R_x  = U1_holder[:,:m1].dot(x)
        R_x += U2_holder[:,:m1].dot(y)
        R_x -= X_full * w
        R_y  = U2_holder[:,:m1].dot(x)
        R_y += U1_holder[:,:m1].dot(y)
        R_y += Y_full * w

        r_norms  = np.linalg.norm(R_x, axis=0) ** 2
        r_norms += np.linalg.norm(R_y, axis=0) ** 2
        r_norms = r_norms ** .5

        if x0sym is not None:
            xt_ir = v_ir[:space_inc]
            x0_ir = v_ir[:nroots]

        max_r_norm = max(r_norms[:nroots])
        conv = r_norms[:nroots] <= tol_residual
        for k, ek in enumerate(e[:nroots]):
            if conv[k] and not conv_last[k]:
                log.debug('root %d converged  |r|= %4.3g  e= %s  max|de|= %4.3g',
                          k, r_norms[k], ek, de[k])
            else:
                log.debug1('root %d  |r|= %4.3g  e= %s  max|de|= %4.3g',
                          k, r_norms[k], ek, de[k])
        ide = np.argmax(abs(de))
        if all(conv):
            log.debug('converged %d %d  |r|= %4.3g  e= %s  max|de|= %4.3g',
                      icyc, len(conv), max_r_norm, e, de[ide])
            break

        r_index = r_norms > tol_residual
        XY_new = precond(np.vstack([R_x[:,r_index], R_y[:,r_index]]).T, w[r_index])
        X_new = XY_new[:,:A_size].T
        Y_new = XY_new[:,A_size:].T
        if x0sym is None:
            V, W = VW_Gram_Schmidt_fill_holder(
                V_holder[:,:m1], W_holder[:,:m1], X_new, Y_new, lindep)
        else:
            xt_ir = xt_ir[r_index]
            xt_orth_ir = []
            V = []
            W = []
            for ir in set(xt_ir):
                idx = np.nonzero(xt_ir == ir)[0]
                _V, _W = VW_Gram_Schmidt_fill_holder(
                    V_holder[:,:m1], W_holder[:,:m1], X_new[:,idx], Y_new[:,idx], lindep)
                V.append(_V)
                W.append(_W)
                xt_orth_ir.append([ir] * len(_V))
            if len(V) > 0:
                V = np.vstack(V)
                W = np.vstack(W)
                xt_ir = np.hstack(xt_orth_ir)
                xs_ir = np.hstack([xs_ir, xt_ir])

        if len(V) == 0:
            log.debug(f'Linear dependency in trial subspace. |r| for each state {r_norms}')
            break

        log.debug1('Generate %d trial vectors. Drop %d vectors',
                   len(V), r_norms.size - len(V))
        X_new = Y_new = R_x = R_y = None

        xy_norms  = np.linalg.norm(V, axis=0) ** 2
        xy_norms += np.linalg.norm(W, axis=0) ** 2
        norm_min = (xy_norms ** .5).min()
        log.debug('real_lr_eig %d %d  |r|= %4.3g  e= %s  max|de|= %4.3g  lindep= %4.3g',
                  icyc, m1, max_r_norm, e, de[ide], norm_min)

        fresh_start = m1 + len(V) > max_space

    # Check whether the solver finds enough eigenvectors.
    if len(x0[0]) < min(A_size, nroots):
        log.warn(f'Not enough eigenvectors (len(x0)={len(x0[0])}, nroots={nroots})')

    return conv[:nroots], e[:nroots], np.hstack(x0)

def _gen_x0(v, xs):
    out = _outprod_to_subspace(v[::2], xs)
    out_conj = _outprod_to_subspace(v[1::2].conj(), xs)
    n = out.shape[1] // 2
    # v[1::2] * xs.conj() = (v[1::2].conj() * xs).conj()
    out[:,:n] += out_conj[:,n:].conj()
    out[:,n:] += out_conj[:,:n].conj()
    return out

def _gen_ax0(v, xs):
    out = _outprod_to_subspace(v[::2], xs)
    out_conj = _outprod_to_subspace(v[1::2].conj(), xs)
    n = out.shape[1] // 2
    out[:,:n] -= out_conj[:,n:].conj()
    out[:,n:] -= out_conj[:,:n].conj()
    return out

def _conj_dot(xi, xj):
    '''Dot product between the conjugated basis of xi and xj.
    The conjugated basis of xi is np.hstack([xi[half:], xi[:half]]).conj()
    '''
    n = xi.shape[-1] // 2
    return xi[:,n:].dot(xj[:,:n].T) + xi[:,:n].dot(xj[:,n:].T)

def _qr(xs, lindep=1e-14):
    '''QR decomposition for a list of vectors (for linearly independent vectors only).
    xs = (r.T).dot(qs)
    '''
    nv = 0
    idx = []
    for i, xi in enumerate(xs):
        for j in range(nv):
            prod = xs[j].conj().dot(xi)
            xi -= xs[j] * prod
        norm = np.linalg.norm(xi)
        if norm**2 > lindep:
            xs[nv] = xi/norm
            nv += 1
            idx.append(i)
    return xs[:nv], idx

def _symmetric_orth(xt, lindep=1e-6):
    xt = np.asarray(xt)
    if xt.dtype == np.float64:
        return _symmetric_orth_real(xt, lindep)
    else:
        return _symmetric_orth_cmplx(xt, lindep)

def _symmetric_orth_real(xt, lindep=1e-6):
    '''
    Symmetric orthogonalization for xt = {[X, Y]},
    and its dual basis vectors {[Y, X]}
    '''
    x0_size = xt.shape[1]
    s11 = xt.dot(xt.T)
    s21 = _conj_dot(xt, xt)
    # Symmetric orthogonalize s, where
    # s = [[s11, s21.conj().T],
    #      [s21, s11.conj()  ]]
    e, c = np.linalg.eigh(s11)
    mask = e > lindep**2
    e = e[mask]
    if e.size == 0:
        return np.zeros([0, x0_size], dtype=xt.dtype)
    c = c[:,mask] * e**-.5

    # c22 = c.conj()
    # s21 -> c22.conj().T.dot(s21).dot(c11)
    csc = c.T.dot(s21).dot(c)
    n = csc.shape[0]
    for i in range(n):
        _s21 = csc[i:,i:]
        # s21 is symmetric for real vectors
        w, u = np.linalg.eigh(_s21)
        mask = 1 - abs(w) > lindep
        if np.any(mask):
            c = c[:,i:]
            break
    else:
        return np.zeros([0, x0_size], dtype=xt.dtype)
    w = w[mask]
    u = u[:,mask]
    c_orth = c.dot(u)

    if e[0] < 1e-6 or 1-abs(w[0]) < 1e-3:
        # Rerun the orthogonalization to reduce numerical errors
        e, c = np.linalg.eigh(c_orth.T.dot(s11).dot(c_orth))
        c *= e**-.5
        c_orth = c_orth.dot(c)
        csc = c_orth.T.dot(s21).dot(c_orth)
        w, u = np.linalg.eigh(csc)
        c_orth = c_orth.dot(u)

    # Symmetric diagonalize
    # [1 w.conj()] => c = [a b]
    # [w 1       ]        [b a]
    # where
    # a = ((1+w)**-.5 + (1-w)**-.5)/2
    # b = (phase*(1+w)**-.5 - phase*(1-w)**-.5)/2
    a1 = (1 + w)**-.5
    a2 = (1 - w)**-.5
    a = (a1 + a2) / 2
    b = (a1 - a2) / 2

    m = xt.shape[1] // 2
    x_orth = (c_orth * a).T.dot(xt)
    # Contribution from the conjugated basis
    x_orth[:,:m] += (c_orth * b).T.dot(xt[:,m:])
    x_orth[:,m:] += (c_orth * b).T.dot(xt[:,:m])
    return x_orth

def _symmetric_orth_cmplx(xt, lindep=1e-6):
    n, m = xt.shape
    if n == 0:
        raise RuntimeError('Linear dependency in trial bases')
    m = m // 2
    # The conjugated basis np.hstack([xt[:,m:], xt[:,:m]]).conj()
    s11 = xt.conj().dot(xt.T)
    s21 = _conj_dot(xt, xt)
    s = np.block([[s11, s21.conj().T],
                  [s21, s11.conj()  ]])
    e, c = scipy.linalg.eigh(s)
    if e[0] < lindep:
        if n == 1:
            return xt
        return _symmetric_orth_cmplx(xt[:-1], lindep)

    c_orth = (c * e**-.5).dot(c[:n].conj().T)
    x_orth = c_orth[:n].T.dot(xt)
    # Contribution from the conjugated basis
    x_orth[:,:m] += c_orth[n:].T.dot(xt[:,m:].conj())
    x_orth[:,m:] += c_orth[n:].T.dot(xt[:,:m].conj())
    return x_orth

def _sym_dot(V, U1, m0, m1):
    '''(V*U1 + U1.T*V.T)[m0:m1,:m1]'''
    a  = V [:,m0:m1].T.dot(U1[:,:m1])
    a += U1[:,m0:m1].T.dot(V [:,:m1])
    a *= .5
    return a

def _asym_dot(V, U1, m0, m1):
    '''(V*U1 - U1.T*V.T)[m0:m1,:m1]'''
    a  = V [:,m0:m1].T.dot(U1[:,:m1])
    a -= U1[:,m0:m1].T.dot(V [:,:m1])
    a *= .5
    return a

def TDDFT_subspace_eigen_solver(a, b, sigma, pi, nroots):
    ''' [ a b ] x - [ σ   π] x  Ω = 0 '''
    ''' [ b a ] y   [-π  -σ] y    = 0 '''

    d = abs(np.diag(sigma))
    d_mh = d**(-0.5)

    s_m_p = np.einsum('i,ij,j->ij', d_mh, sigma - pi, d_mh)

    '''LU = d^−1/2 (σ − π) d^−1/2'''
    ''' A = LU '''
    L, U = scipy.linalg.lu(s_m_p, permute_l=True)
    L_inv = np.linalg.inv(L)
    U_inv = np.linalg.inv(U)

    '''U^-T d^−1/2 (a−b) d^-1/2 U^-1 = GG^T '''
    d_amb_d = np.einsum('i,ij,j->ij', d_mh, a-b, d_mh)
    GGT = np.linalg.multi_dot([U_inv.T, d_amb_d, U_inv])

    G = scipy.linalg.cholesky(GGT, lower=True)
    G_inv = np.linalg.inv(G)

    ''' M = G^T L^−1 d^−1/2 (a+b) d^−1/2 L^−T G '''
    d_apb_d = np.einsum('i,ij,j->ij', d_mh, a+b, d_mh)
    M = np.linalg.multi_dot([G.T, L_inv, d_apb_d, L_inv.T, G])

    omega2, Z = np.linalg.eigh(M)
    if np.any(omega2 <= 0):
        idx = np.nonzero(omega2 > 0)[0]
        omega2 = omega2[idx[:nroots]]
        Z = Z[:,idx[:nroots]]
    else:
        omega2 = omega2[:nroots]
        Z = Z[:,:nroots]
    omega = omega2**0.5

    ''' It requires Z^T Z = 1/Ω '''
    ''' x+y = d^−1/2 L^−T GZ Ω^-0.5 '''
    ''' x−y = d^−1/2 U^−1 G^−T Z Ω^0.5 '''
    x_p_y = np.einsum('i,ik,k->ik', d_mh, L_inv.T.dot(G.dot(Z)), omega**-0.5)
    x_m_y = np.einsum('i,ik,k->ik', d_mh, U_inv.dot(G_inv.T.dot(Z)), omega**0.5)

    x = (x_p_y + x_m_y)/2
    y = x_p_y - x
    return omega, x, y

def VW_Gram_Schmidt_fill_holder(V_holder, W_holder, X_new, Y_new, lindep=1e-12):
    '''
    QR orthogonalization for (X_new, Y_new) basis vectors, then apply symmetric
    orthogonalization for {[X, Y]}, and its dual basis vectors {[Y, X]}
    '''
    _x  = V_holder.T.dot(X_new)
    _x += W_holder.T.dot(Y_new)
    _y  = V_holder.T.dot(Y_new)
    _y += W_holder.T.dot(X_new)
    X_new -= V_holder.dot(_x)
    X_new -= W_holder.dot(_y)
    Y_new -= W_holder.dot(_x)
    Y_new -= V_holder.dot(_y)
    x0_size = X_new.shape[0]

    s11  = X_new.T.dot(X_new)
    s11 += Y_new.T.dot(Y_new)
    # s21 is symmetric
    s21  = X_new.T.dot(Y_new)
    s21 += Y_new.T.dot(X_new)
    e, c = np.linalg.eigh(s11)
    mask = e > lindep
    e = e[mask]
    if e.size == 0:
        return (np.zeros([0, x0_size], dtype=X_new.dtype),
                np.zeros([0, x0_size], dtype=Y_new.dtype))
    c = c[:,mask] * e**-.5

    csc = c.T.dot(s21).dot(c)
    n = csc.shape[0]
    lindep_sqrt = lindep**.5
    for i in range(n):
        w, u = np.linalg.eigh(csc[i:,i:])
        mask = 1 - abs(w) > lindep_sqrt
        if np.any(mask):
            c = c[:,i:]
            break
    else:
        return (np.zeros([0, x0_size], dtype=X_new.dtype),
                np.zeros([0, x0_size], dtype=Y_new.dtype))
    w = w[mask]
    u = u[:,mask]
    c_orth = c.dot(u)

    if e[0] < lindep_sqrt or any(abs(w)> 1-1e-3):
        # Rerun the orthogonalization to reduce numerical errors.
        # When w~=1-1e-3, errors in the orthogonalization (off-diagonal terms)
        # is near 1e-6.
        e, c = np.linalg.eigh(c_orth.T.dot(s11).dot(c_orth))
        c *= e**-.5
        c_orth = c_orth.dot(c)
        csc = c_orth.T.dot(s21).dot(c_orth)
        w, u = np.linalg.eigh(csc)

        mask = 1 - abs(w) > lindep_sqrt
        w = w[mask]
        u = u[:,mask]
        c_orth = c_orth.dot(u)

    # Symmetric diagonalize
    # [1 w] => c = [a b]
    # [w 1]        [b a]
    # where
    # a = ((1+w)**-.5 + (1-w)**-.5)/2
    # b = ((1+w)**-.5 - (1-w)**-.5)/2
    a1 = (1 + w)**-.5
    a2 = (1 - w)**-.5
    a = (a1 + a2) / 2
    b = (a1 - a2) / 2

    x_orth  = X_new.dot(c_orth * a)
    x_orth += Y_new.dot(c_orth * b)
    y_orth  = Y_new.dot(c_orth * a)
    y_orth += X_new.dot(c_orth * b)
    return x_orth.T, y_orth.T



#!/usr/bin/env python
# Copyright 2014-2019 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
QM/MM helper functions that modify the QM methods.
'''

import numpy
import pyscf
from pyscf import lib
from pyscf import gto
from pyscf import df
from pyscf import scf
from pyscf import mcscf
from pyscf import grad
from pyscf.lib import logger
from pyscf.qmmm import mm_mole


def add_mm_charges(scf_method, atoms_or_coords, charges, radii=None, unit=None):
    '''Embedding the one-electron (non-relativistic) potential generated by MM
    point charges into QM Hamiltonian.

    The total energy includes the regular QM energy, the interaction between
    the nuclei in QM region and the MM charges, and the static Coulomb
    interaction between the electron density and the MM charges. It does not
    include the static Coulomb interactions of the MM point charges, the MM
    energy, the vdw interaction or other bonding/non-bonding effects between
    QM region and MM particles.

    Args:
        scf_method : a HF or DFT object

        atoms_or_coords : 2D array, shape (N,3)
            MM particle coordinates
        charges : 1D array
            MM particle charges

    Kwargs:
        radii : 1D array
            The Gaussian charge distribution radii of MM atoms.
        unit : str
            Bohr, AU, Ang (case insensitive). Default is the same to mol.unit

    Returns:
        Same method object as the input scf_method with modified 1e Hamiltonian

    Note:
        1. if MM charge and X2C correction are used together, function mm_charge
        needs to be applied after X2C decoration (.x2c method), eg
        mf = mm_charge(scf.RHF(mol).x2c()), [(0.5,0.6,0.8)], [-0.5]).
        2. Once mm_charge function is applied on the SCF object, it
        affects all the post-HF calculations eg MP2, CCSD, MCSCF etc

    Examples:

    >>> mol = gto.M(atom='H 0 0 0; F 0 0 1', basis='ccpvdz', verbose=0)
    >>> mf = mm_charge(dft.RKS(mol), [(0.5,0.6,0.8)], [-0.3])
    >>> mf.kernel()
    -101.940495711284
    '''
    mol = scf_method.mol
    if unit is None:
        unit = mol.unit
    mm_mol = mm_mole.create_mm_mol(atoms_or_coords, charges,
                                   radii=radii, unit=unit)
    return qmmm_for_scf(scf_method, mm_mol)

# Define method mm_charge for backward compatibility
mm_charge = add_mm_charges

def qmmm_for_scf(method, mm_mol):
    '''Add the potential of MM particles to SCF (HF and DFT) method or CASCI
    method then generate the corresponding QM/MM method for the QM system.

    Args:
        mm_mol : MM Mole object
    '''
    assert (isinstance(method, (scf.hf.SCF, mcscf.casci.CASBase)))

    if isinstance(method, scf.hf.SCF):
        # Avoid to initialize QMMM twice
        if isinstance(method, QMMM):
            method.mm_mol = mm_mol
            return method

        cls = QMMMSCF
    else:
        # post-HF methods
        if isinstance(method._scf, QMMM):
            method._scf.mm_mol = mm_mol
            return method

        cls = QMMMPostSCF

    return lib.set_class(cls(method, mm_mol), (cls, method.__class__))

class QMMM:
    __name_mixin__ = 'QMMM'

_QMMM = QMMM

class QMMMSCF(QMMM):
    _keys = {'mm_mol'}

    def __init__(self, method, mm_mol=None):
        self.__dict__.update(method.__dict__)
        if mm_mol is None:
            mm_mol = gto.Mole()
        self.mm_mol = mm_mol

    def undo_qmmm(self):
        obj = lib.view(self, lib.drop_class(self.__class__, QMMM))
        del obj.mm_mol
        return obj

    def dump_flags(self, verbose=None):
        super().dump_flags(verbose)
        logger.info(self, '** Add background charges for %s **',
                    self.__class__.__name__)
        if self.verbose >= logger.DEBUG:
            logger.debug(self, 'Charge      Location')
            coords = self.mm_mol.atom_coords()
            charges = self.mm_mol.atom_charges()
            for i, z in enumerate(charges):
                logger.debug(self, '%.9g    %s', z, coords[i])
        return self

    def get_hcore(self, mol=None):
        if mol is None:
            mol = self.mol
        mm_mol = self.mm_mol

        h1e = super().get_hcore(mol)

        coords = mm_mol.atom_coords()
        charges = mm_mol.atom_charges()
        nao = mol.nao
        max_memory = self.max_memory - lib.current_memory()[0]
        blksize = int(min(max_memory*1e6/8/nao**2, 200))
        blksize = max(blksize, 1)
        if mm_mol.charge_model == 'gaussian':
            expnts = mm_mol.get_zetas()

            if mol.cart:
                intor = 'int3c2e_cart'
            else:
                intor = 'int3c2e_sph'
            cintopt = gto.moleintor.make_cintopt(mol._atm, mol._bas,
                                                 mol._env, intor)
            v = 0
            for i0, i1 in lib.prange(0, charges.size, blksize):
                fakemol = gto.fakemol_for_charges(coords[i0:i1], expnts[i0:i1])
                j3c = df.incore.aux_e2(mol, fakemol, intor=intor,
                                       aosym='s2ij', cintopt=cintopt)
                v += numpy.einsum('xk,k->x', j3c, -charges[i0:i1])
            v = lib.unpack_tril(v)
            h1e += v
        else:
            for i0, i1 in lib.prange(0, charges.size, blksize):
                j3c = mol.intor('int1e_grids', hermi=1, grids=coords[i0:i1])
                h1e += numpy.einsum('kpq,k->pq', j3c, -charges[i0:i1])
        return h1e

    def energy_nuc(self):
        # interactions between QM nuclei and MM particles
        nuc = self.mol.energy_nuc()
        coords = self.mm_mol.atom_coords()
        charges = self.mm_mol.atom_charges()
        for j in range(self.mol.natm):
            q2, r2 = self.mol.atom_charge(j), self.mol.atom_coord(j)
            r = lib.norm(r2-coords, axis=1)
            nuc += q2*(charges/r).sum()
        return nuc

    def to_gpu(self):
        obj = self.undo_qmmm().to_gpu()
        obj = qmmm_for_scf(obj, self.mm_mol)
        return lib.to_gpu(self, obj)

    def nuc_grad_method(self):
        scf_grad = super().nuc_grad_method()
        return qmmm_grad_for_scf(scf_grad)

    Gradients = nuc_grad_method

class QMMMPostSCF(QMMM):
    def __init__(self, method, mm_mol=None):
        self.__dict__.update(method.__dict__)
        mf = qmmm_for_scf(method._scf, mm_mol).run()
        self._scf = mf
        self.mo_coeff = mf.mo_coeff
        self.mo_energy = mf.mo_energy

    def undo_qmmm(self):
        obj = lib.view(self, lib.drop_class(self.__class__, QMMM))
        obj._scf = self._scf.undo_qmmm()
        return obj

    to_gpu = QMMMSCF.to_gpu


def add_mm_charges_grad(scf_grad, atoms_or_coords, charges, radii=None, unit=None):
    '''Apply the MM charges in the QM gradients' method.  It affects both the
    electronic and nuclear parts of the QM fragment.

    Args:
        scf_grad : a HF or DFT gradient object (grad.HF or grad.RKS etc)
            Once the add_mm_charges_grad was applied, it affects all post-HF
            calculations eg MP2, CCSD, MCSCF etc
        coords : 2D array, shape (N,3)
            MM particle coordinates
        charges : 1D array
            MM particle charges
    Kwargs:
        radii : 1D array
            The Gaussian charge distribution radii of MM atoms.
        unit : str
            Bohr, AU, Ang (case insensitive). Default is the same to mol.unit

    Returns:
        Same gradeints method object as the input scf_grad method

    Examples:

    >>> from pyscf import gto, scf, grad
    >>> mol = gto.M(atom='H 0 0 0; F 0 0 1', basis='ccpvdz', verbose=0)
    >>> mf = mm_charge(scf.RHF(mol), [(0.5,0.6,0.8)], [-0.3])
    >>> mf.kernel()
    -101.940495711284
    >>> hfg = mm_charge_grad(grad.hf.RHF(mf), coords, charges)
    >>> hfg.kernel()
    [[-0.25912357 -0.29235976 -0.38245077]
     [-1.70497052 -1.89423883  1.2794798 ]]
    '''
    assert (isinstance(scf_grad, grad.rhf.Gradients))
    mol = scf_grad.mol
    if unit is None:
        unit = mol.unit
    mm_mol = mm_mole.create_mm_mol(atoms_or_coords, charges,
                                   radii=radii, unit=unit)
    mm_grad = qmmm_grad_for_scf(scf_grad)
    mm_grad.base.mm_mol = mm_mol
    return mm_grad

# Define method mm_charge_grad for backward compatibility
mm_charge_grad = add_mm_charges_grad

def qmmm_grad_for_scf(scf_grad):
    '''Add the potential of MM particles to SCF (HF and DFT) object and then
    generate the corresponding QM/MM gradients method for the QM system.
    '''
    if getattr(scf_grad.base, 'with_x2c', None):
        raise NotImplementedError('X2C with QM/MM charges')

    # Avoid to initialize QMMMGrad twice
    if isinstance(scf_grad, QMMMGrad):
        return scf_grad

    assert (isinstance(scf_grad.base, scf.hf.SCF) and
           isinstance(scf_grad.base, QMMM))

    return scf_grad.view(lib.make_class((QMMMGrad, scf_grad.__class__)))

class QMMMGrad:
    __name_mixin__ = 'QMMM'

    def __init__(self, scf_grad):
        self.__dict__.update(scf_grad.__dict__)

    def dump_flags(self, verbose=None):
        super().dump_flags(verbose)
        logger.info(self, '** Add background charges for %s **',
                    self.__class__.__name__)
        if self.verbose >= logger.DEBUG1:
            logger.debug1(self, 'Charge      Location')
            coords = self.base.mm_mol.atom_coords()
            charges = self.base.mm_mol.atom_charges()
            for i, z in enumerate(charges):
                logger.debug1(self, '%.9g    %s', z, coords[i])
        return self

    def get_hcore(self, mol=None):
        ''' (QM 1e grad) + <-d/dX i|q_mm/r_mm|j>'''
        if mol is None:
            mol = self.mol
        mm_mol = self.base.mm_mol
        coords = mm_mol.atom_coords()
        charges = mm_mol.atom_charges()

        nao = mol.nao
        max_memory = self.max_memory - lib.current_memory()[0]
        blksize = int(min(max_memory*1e6/8/nao**2/3, 200))
        blksize = max(blksize, 1)
        g_qm = super().get_hcore(mol)
        if mm_mol.charge_model == 'gaussian':
            expnts = mm_mol.get_zetas()
            if mol.cart:
                intor = 'int3c2e_ip1_cart'
            else:
                intor = 'int3c2e_ip1_sph'
            cintopt = gto.moleintor.make_cintopt(mol._atm, mol._bas,
                                                 mol._env, intor)
            v = 0
            for i0, i1 in lib.prange(0, charges.size, blksize):
                fakemol = gto.fakemol_for_charges(coords[i0:i1], expnts[i0:i1])
                j3c = df.incore.aux_e2(mol, fakemol, intor, aosym='s1',
                                       comp=3, cintopt=cintopt)
                v += numpy.einsum('ipqk,k->ipq', j3c, charges[i0:i1])
            g_qm += v
        else:
            for i0, i1 in lib.prange(0, charges.size, blksize):
                j3c = mol.intor('int1e_grids_ip', grids=coords[i0:i1])
                g_qm += numpy.einsum('ikpq,k->ipq', j3c, charges[i0:i1])
        return g_qm

    def grad_hcore_mm(self, dm, mol=None):
        r'''Nuclear gradients of the electronic energy
        with respect to MM atoms:

        ... math::
            g = \sum_{ij} \frac{\partial hcore_{ij}}{\partial R_{I}} P_{ji},

        where I represents MM atoms.

        Args:
            dm : array
                The QM density matrix.
        '''
        if mol is None:
            mol = self.mol
        mm_mol = self.base.mm_mol

        coords = mm_mol.atom_coords()
        charges = mm_mol.atom_charges()
        expnts = mm_mol.get_zetas()

        intor = 'int3c2e_ip2'
        nao = mol.nao
        max_memory = self.max_memory - lib.current_memory()[0]
        blksize = int(min(max_memory*1e6/8/nao**2/3, 200))
        blksize = max(blksize, 1)
        cintopt = gto.moleintor.make_cintopt(mol._atm, mol._bas,
                                             mol._env, intor)

        g = numpy.empty_like(coords)
        for i0, i1 in lib.prange(0, charges.size, blksize):
            fakemol = gto.fakemol_for_charges(coords[i0:i1], expnts[i0:i1])
            j3c = df.incore.aux_e2(mol, fakemol, intor, aosym='s1',
                                   comp=3, cintopt=cintopt)
            g[i0:i1] = numpy.einsum('ipqk,qp->ik', j3c * charges[i0:i1], dm).T
        return g

    contract_hcore_mm = grad_hcore_mm # for backward compatibility

    def grad_nuc(self, mol=None, atmlst=None):
        if mol is None: mol = self.mol
        coords = self.base.mm_mol.atom_coords()
        charges = self.base.mm_mol.atom_charges()

        g_qm = super().grad_nuc(mol, atmlst)
# nuclei lattice interaction
        g_mm = numpy.empty((mol.natm,3))
        for i in range(mol.natm):
            q1 = mol.atom_charge(i)
            r1 = mol.atom_coord(i)
            r = lib.norm(r1-coords, axis=1)
            g_mm[i] = -q1 * numpy.einsum('i,ix,i->x', charges, r1-coords, 1/r**3)
        if atmlst is not None:
            g_mm = g_mm[atmlst]
        return g_qm + g_mm

    def grad_nuc_mm(self, mol=None):
        '''Nuclear gradients of the QM-MM nuclear energy
        (in the form of point charge Coulomb interactions)
        with respect to MM atoms.
        '''
        if mol is None:
            mol = self.mol
        mm_mol = self.base.mm_mol
        coords = mm_mol.atom_coords()
        charges = mm_mol.atom_charges()
        g_mm = numpy.zeros_like(coords)
        for i in range(mol.natm):
            q1 = mol.atom_charge(i)
            r1 = mol.atom_coord(i)
            r = lib.norm(r1-coords, axis=1)
            g_mm += q1 * numpy.einsum('i,ix,i->ix', charges, r1-coords, 1/r**3)
        return g_mm

    def to_gpu(self):
        obj = self.undo_qmmm().to_gpu()
        obj = qmmm_grad_for_scf(obj)
        return lib.to_gpu(self, obj)

_QMMMGrad = QMMMGrad

# Inject QMMM interface wrapper to other modules
scf.hf.SCF.QMMM = mm_charge
mcscf.casci.CASBase.QMMM = mm_charge
grad.rhf.Gradients.QMMM = mm_charge_grad

if __name__ == '__main__':
    from pyscf import scf, cc, grad
    mol = gto.Mole()
    mol.atom = ''' O                  0.00000000    0.00000000   -0.11081188
                   H                 -0.00000000   -0.84695236    0.59109389
                   H                 -0.00000000    0.89830571    0.52404783 '''
    mol.basis = 'cc-pvdz'
    mol.build()

    coords = [(0.5,0.6,0.8)]
    #coords = [(0.0,0.0,0.0)]
    charges = [-0.5]
    mf = mm_charge(scf.RHF(mol), coords, charges)
    print(mf.kernel()) # -76.3206550372

    g = mf.nuc_grad_method().kernel()
    mfs = mf.as_scanner()
    e1 = mfs(''' O                  0.00100000    0.00000000   -0.11081188
             H                 -0.00000000   -0.84695236    0.59109389
             H                 -0.00000000    0.89830571    0.52404783 ''')
    e2 = mfs(''' O                 -0.00100000    0.00000000   -0.11081188
             H                 -0.00000000   -0.84695236    0.59109389
             H                 -0.00000000    0.89830571    0.52404783 ''')
    print((e1 - e2)/0.002 * lib.param.BOHR, g[0,0])

    mycc = cc.ccsd.CCSD(mf)
    ecc, t1, t2 = mycc.kernel() # ecc = -0.228939687075

    g = mycc.nuc_grad_method().kernel()
    ccs = mycc.as_scanner()
    e1 = ccs(''' O                  0.00100000    0.00000000   -0.11081188
             H                 -0.00000000   -0.84695236    0.59109389
             H                 -0.00000000    0.89830571    0.52404783 ''')
    e2 = ccs(''' O                 -0.00100000    0.00000000   -0.11081188
             H                 -0.00000000   -0.84695236    0.59109389
             H                 -0.00000000    0.89830571    0.52404783 ''')
    print((e1 - e2)/0.002 * lib.param.BOHR, g[0,0])
